@@ -26,7 +26,7 @@
    Boston, MA 02111 USA.
 
    <title>GSObjCRuntime function and macro reference</title>
-   $Date: 2011-10-15 12:36:51 -0600 (Sat, 15 Oct 2011) $ $Revision: 34005 $
+   $Date: 2013-11-22 15:24:26 +0000 (Fri, 22 Nov 2013) $ $Revision: 37395 $
    */
 
 #import "common.h"
@@ -45,13 +45,13 @@
 #import "Foundation/NSValue.h"
 #endif
 #import "GNUstepBase/GSObjCRuntime.h"
+#import "GNUstepBase/NSObject+GNUstepBase.h"
 
 #import "../GSPrivate.h"
 
 #include <objc/Protocol.h>
 
 #include <stdio.h>
-#include <string.h>
 #include <ctype.h>
 
 #ifndef NeXT_RUNTIME
@@ -444,8 +444,6 @@ GSObjCMakeClass(NSString *name, NSString *superName, NSDictionary *iVars)
   Class		newClass;
   Class		classSuperClass;
   const char	*classNameCString;
-  char		*tmp;
-  int		len;
 
   NSCAssert(name, @"no name");
   NSCAssert(superName, @"no superName");
@@ -456,12 +454,6 @@ GSObjCMakeClass(NSString *name, NSString *superName, NSDictionary *iVars)
   NSCAssert1(!NSClassFromString(name), @"A class %@ already exists", name);
 
   classNameCString = [name UTF8String];
-  len = strlen(classNameCString);
-  tmp = malloc(len + 1);
-  strncpy(tmp, classNameCString, len);
-  tmp[len] = '\0';
-  classNameCString = tmp;
-
   newClass = objc_allocateClassPair(classSuperClass, classNameCString, 0);
   if ([iVars count] > 0)
     {
@@ -627,8 +619,8 @@ GSGetMethod(Class cls, SEL sel,
 }
 
 
-static inline const char *
-gs_skip_type_qualifier_and_layout_info (const char *types)
+GS_EXPORT const char *
+GSSkipTypeQualifierAndLayoutInfo(const char *types)
 {
   while (*types == '+'
     || *types == '-'
@@ -653,38 +645,49 @@ GS_EXPORT BOOL
 GSSelectorTypesMatch(const char *types1, const char *types2)
 {
   if (! types1 || ! types2)
-    return NO;
-
+    {
+      return NO;        // Nul pointers never match
+    }
+  if (types1 == types2)
+    {
+      return YES;
+    }
   while (*types1 && *types2)
     {
-      types1 = gs_skip_type_qualifier_and_layout_info (types1);
-      types2 = gs_skip_type_qualifier_and_layout_info (types2);
+      types1 = GSSkipTypeQualifierAndLayoutInfo (types1);
+      types2 = GSSkipTypeQualifierAndLayoutInfo (types2);
 
       /* Reached the end of the selector.  */
       if (! *types1 && ! *types2)
-        return YES;
+        {
+          return YES;
+        }
 
       /* Ignore structure name yet compare layout.  */
       if (*types1 == '{' && *types2 == '{')
 	{
 	  while (*types1 != '=' && *types1 != '}')
-	    types1++;
-
+            {
+              types1++;
+            }
 	  while (*types2 != '=' && *types2 != '}')
-	    types2++;
+            {
+              types2++;
+            }
 	}
 
       if (*types1 != *types2)
-        return NO;
-
+        {
+          return NO;
+        }
       types1++;
       types2++;
     }
 
-  types1 = gs_skip_type_qualifier_and_layout_info (types1);
-  types2 = gs_skip_type_qualifier_and_layout_info (types2);
+  types1 = GSSkipTypeQualifierAndLayoutInfo (types1);
+  types2 = GSSkipTypeQualifierAndLayoutInfo (types2);
 
-  return (! *types1 && ! *types2);
+  return (! *types1 && ! *types2) ? YES : NO;
 }
 
 /* See header for documentation. */
@@ -826,16 +829,19 @@ GSProtocolGetMethodDescriptionRecursive(Protocol *aProtocol, SEL aSel, BOOL isRe
     {
       Protocol **list;
       unsigned int count;
+
       list = protocol_copyProtocolList(aProtocol, &count);
       if (list != NULL)
         {
           unsigned int i;
+
           for (i = 0; i < count; i++)
             {
-              desc = GSProtocolGetMethodDescriptionRecursive(list[i], aSel, isRequired, isInstance);
+              desc = GSProtocolGetMethodDescriptionRecursive(list[i],
+                aSel, isRequired, isInstance);
               if (desc.name != NULL || desc.types != NULL)
                 {
-                  return desc;
+                  break;
                 }
             }
           free(list);
@@ -1293,7 +1299,7 @@ GSObjCGetVal(NSObject *self, const char *key, SEL sel,
             break;
 
           case _C_STRUCT_B:
-            if (strcmp(@encode(NSPoint), type) == 0)
+            if (GSSelectorTypesMatch(@encode(NSPoint), type))
               {
                 NSPoint	v;
 
@@ -1310,7 +1316,7 @@ GSObjCGetVal(NSObject *self, const char *key, SEL sel,
                   }
                 val = [NSValue valueWithPoint: v];
               }
-            else if (strcmp(@encode(NSRange), type) == 0)
+            else if (GSSelectorTypesMatch(@encode(NSRange), type))
               {
                 NSRange	v;
 
@@ -1327,7 +1333,7 @@ GSObjCGetVal(NSObject *self, const char *key, SEL sel,
                   }
                 val = [NSValue valueWithRange: v];
               }
-            else if (strcmp(@encode(NSRect), type) == 0)
+            else if (GSSelectorTypesMatch(@encode(NSRect), type))
               {
                 NSRect	v;
 
@@ -1344,7 +1350,7 @@ GSObjCGetVal(NSObject *self, const char *key, SEL sel,
                   }
                 val = [NSValue valueWithRect: v];
               }
-            else if (strcmp(@encode(NSSize), type) == 0)
+            else if (GSSelectorTypesMatch(@encode(NSSize), type))
               {
                 NSSize	v;
 
@@ -1397,8 +1403,7 @@ GSObjCGetVal(NSObject *self, const char *key, SEL sel,
 
 	      cls = [self class];
 	      type_slot = objc_get_slot(cls, @selector(retain));
-	      typed = GSSelectorFromNameAndTypes(sel_getName(sel),
-		type_slot->types);
+	      typed = GSSelectorFromNameAndTypes(sel_getName(sel), NULL);
 	      slot = objc_get_slot(cls, typed);
 	      if (strcmp(slot->types, type_slot->types) == 0)
 		{
@@ -1740,7 +1745,7 @@ GSObjCSetVal(NSObject *self, const char *key, id val, SEL sel,
 	    break;
 
           case _C_STRUCT_B:
-            if (strcmp(@encode(NSPoint), type) == 0)
+            if (GSSelectorTypesMatch(@encode(NSPoint), type))
               {
                 NSPoint	v = [val pointValue];
 
@@ -1758,7 +1763,7 @@ GSObjCSetVal(NSObject *self, const char *key, id val, SEL sel,
                     (*imp)(self, sel, v);
                   }
               }
-            else if (strcmp(@encode(NSRange), type) == 0)
+            else if (GSSelectorTypesMatch(@encode(NSRange), type))
               {
                 NSRange	v = [val rangeValue];
 
@@ -1776,7 +1781,7 @@ GSObjCSetVal(NSObject *self, const char *key, id val, SEL sel,
                     (*imp)(self, sel, v);
                   }
               }
-            else if (strcmp(@encode(NSRect), type) == 0)
+            else if (GSSelectorTypesMatch(@encode(NSRect), type))
               {
                 NSRect	v = [val rectValue];
 
@@ -1794,7 +1799,7 @@ GSObjCSetVal(NSObject *self, const char *key, id val, SEL sel,
                     (*imp)(self, sel, v);
                   }
               }
-            else if (strcmp(@encode(NSSize), type) == 0)
+            else if (GSSelectorTypesMatch(@encode(NSSize), type))
               {
                 NSSize	v = [val sizeValue];
 
