@@ -1,5 +1,5 @@
 /** Implementation of GNUSTEP printf-style formatting
-   Copyright (C) 1994-2000, 2001 Free Software Foundation, Inc.
+   Copyright (C) 1994-2000, 2001-2013 Free Software Foundation, Inc.
 
    Hacked together by Kai Henningsen <kai@cats.ms>
    from the glibc 2.2.1 sources
@@ -51,14 +51,15 @@
 #endif
 
 #ifdef HAVE_MALLOC_H
+#if !defined(__OpenBSD__)
 #include <malloc.h>
 #endif
+#endif
 #ifdef HAVE_ALLOCA_H
 #include <alloca.h>
 #endif
 
 #include <stdio.h>
-#include <string.h>
 
 #import "Foundation/NSArray.h"
 #import "Foundation/NSCharacterSet.h"
@@ -100,9 +101,6 @@ typedef uint32_t wint_t;
 #ifdef HAVE_SYS_INTTYPES_H
 #include <sys/inttypes.h>
 #endif
-#ifdef HAVE_MALLOC_H
-#include <malloc.h>
-#endif
 #ifndef HAVE_UINTMAX_T
 typedef unsigned long long uintmax_t;
 #endif
@@ -1748,8 +1746,8 @@ NSDictionary *locale)
 	  {
 	    /* This is complicated.  We have to transform the multibyte
 	       string into a unicode string.  */
-	    const char		*str = (const char*)string;
-	    unsigned		blen;
+	    const char			*str = (const char*)string;
+	    unsigned			blen;
 	    static NSStringEncoding	enc = GSUndefinedEncoding;
 	    static BOOL			byteEncoding = NO;
 
@@ -1759,37 +1757,40 @@ NSDictionary *locale)
 		byteEncoding = GSPrivateIsByteEncoding(enc);
 	      }
 
-	    len = strlen(str);	// Number of bytes to convert.
-	    blen = len;		// Size of unichar output buffer.
-
-	    if (prec != -1)
+	    if (-1 == prec)
+              {
+                len = strlen(str);	// Number of bytes to convert.
+                blen = len;		// Size of unichar output buffer.
+              }
+	    else
 	      {
-		if (prec < len)
-		  {
-		    /* We don't neeed an output buffer bigger than the
-		     * precision specifies.
-		     */
-		    blen = prec;
-		  }
 		if (byteEncoding == YES)
 		  {
 		    /* Where the external encoding is one byte per character,
 		     * we know we don't need to convert more bytes than the
 		     * precision required for output.
 		     */
-		    if (prec < len)
-		      {
-			len = prec;
+                    len = 0;
+                    while (len < prec && str[len] != 0)
+                      {
+                        len++;
 		      }
+                    blen = len;
 		  }
-		else if (prec * 4 < len)
+		else
 		  {
-		    /* We assume no multibyte encoding is going to use more
-		     * than the maximum four bytes used by utf-8 for any
-		     * unicode code point.  So we do not need to convert
-		     * more than four times the precision.
-		     */
-		    len = prec * 4;
+                    /* FIXME ... it looks like modern standards mean that
+                     * the number of *bytes* in an input string may not
+                     * exceed the precision ... but that's unintuitive for
+                     * input strings with multibyte characters, so we need
+                     * to check and emulate OSX behavior.
+                     */
+                    len = 0;
+                    while (len < prec && str[len] != 0)
+                      {
+                        len++;
+		      }
+                    blen = len;
 		  }
 	      }
 
@@ -1797,8 +1798,8 @@ NSDictionary *locale)
 	     * enough for the unichar version.
 	     */
 	    if (blen < 8192 || ((string = (unichar *)
-	      NSZoneMalloc(s->_zone, blen * sizeof(unichar))) == NULL))
-	      string = (unichar *) alloca (blen * sizeof(unichar));
+	      NSZoneMalloc(s->_zone, (blen + 1) * sizeof(unichar))) == NULL))
+	      string = (unichar *) alloca((blen + 1) * sizeof(unichar));
 	    else
 	      string_malloced = 1;
 
