@@ -48,8 +48,7 @@
 #import "NSCallBacks.h"
 #import "GSURLPrivate.h"
 #import "GSPrivate.h"
-
-#include <string.h>
+#import "GSTLS.h"
 
 #ifdef	HAVE_SYS_FILE_H
 #  include <sys/file.h>
@@ -228,56 +227,39 @@ static NSLock			*urlLock = nil;
 
 static Class			sslClass = 0;
 
-static NSLock			*debugLock = nil;
-static NSString			*debugFile;
-
-static void debugRead(GSHTTPURLHandle *handle, NSData *data)
+static void
+debugRead(GSHTTPURLHandle *handle, NSData *data)
 {
-  NSString	*s;
-  int		d;
+  int		len = (int)[data length];
+  const char	*ptr = (const char*)[data bytes];
+  int           pos;
 
-  [debugLock lock];
-#if	defined(__MINGW__)
-  d = _wopen((const unichar*)[debugFile  fileSystemRepresentation],
-    O_WRONLY|O_CREAT|O_APPEND, 0644);
-#else
-  d = open([debugFile  fileSystemRepresentation],
-    O_WRONLY|O_CREAT|O_APPEND, 0644);
-#endif
-  if (d >= 0)
+  for (pos = 0; pos < len; pos++)
     {
-      s = [NSString stringWithFormat: @"\nRead for %p at %@ %u bytes - '",
-	handle, [NSDate date], [data length]];
-      write(d, [s cString], [s cStringLength]);
-      write(d, [data bytes], [data length]);
-      write(d, "'", 1);
-      close(d);
+      if (0 == ptr[pos])
+        {
+          NSLog(@"Read for %p of %d bytes - %@", handle, len, data); 
+          return;
+        }
     }
-  [debugLock unlock];
+  NSLog(@"Read for %p of %d bytes - '%*.*s'", handle, len, len, len, ptr); 
 }
-static void debugWrite(GSHTTPURLHandle *handle, NSData *data)
+static void
+debugWrite(GSHTTPURLHandle *handle, NSData *data)
 {
-  NSString	*s;
-  int		d;
+  int		len = (int)[data length];
+  const char	*ptr = (const char*)[data bytes];
+  int           pos = len;
 
-  [debugLock lock];
-#if	defined(__MINGW__)
-  d = _wopen((const unichar*)[debugFile  fileSystemRepresentation],
-    O_WRONLY|O_CREAT|O_APPEND, 0644);
-#else
-  d = open([debugFile  fileSystemRepresentation],
-    O_WRONLY|O_CREAT|O_APPEND, 0644);
-#endif
-  if (d >= 0)
+  for (pos = 0; pos < len; pos++)
     {
-      s = [NSString stringWithFormat: @"\nWrite for %p at %@ %u bytes - '",
-	handle, [NSDate date], [data length]];
-      write(d, [s cString], [s cStringLength]);
-      write(d, [data bytes], [data length]);
-      write(d, "'", 1);
-      close(d);
+      if (0 == ptr[pos])
+        {
+          NSLog(@"Write for %p of %d bytes - %@", handle, len, data); 
+          return;
+        }
     }
-  [debugLock unlock];
+  NSLog(@"Write for %p of %d bytes -'%*.*s'", handle, len, len, len, ptr); 
 }
 
 + (NSURLHandle*) cachedHandleForURL: (NSURL*)newUrl
@@ -308,14 +290,11 @@ static void debugWrite(GSHTTPURLHandle *handle, NSData *data)
   if (self == [GSHTTPURLHandle class])
     {
       urlCache = [NSMutableDictionary new];
+      [[NSObject leakAt: &urlCache] release];
       urlOrder = [NSMutableArray new];
+      [[NSObject leakAt: &urlOrder] release];
       urlLock = [GSLazyLock new];
-      debugLock = [GSLazyLock new];
-      debugFile = [NSString stringWithFormat: @"%@/GSHTTP.%d",
-			     NSTemporaryDirectory(),
-			     [[NSProcessInfo processInfo] processIdentifier]];
-      IF_NO_GC([debugFile retain];)
-
+      [[NSObject leakAt: &urlLock] release];
 #if	!defined(__MINGW__)
       sslClass = [NSFileHandle sslClass];
 #endif
@@ -352,6 +331,7 @@ static void debugWrite(GSHTTPURLHandle *handle, NSData *data)
 {
   if ((self = [super initWithURL: newUrl cached: cached]) != nil)
     {
+      debug = GSDebugSet(@"NSURLHandle");
       dat = [NSMutableData new];
       pageInfo = [NSMutableDictionary new];
       wProperties = NSCreateMapTable(writeKeyCallBacks,
@@ -445,11 +425,16 @@ static void debugWrite(GSHTTPURLHandle *handle, NSData *data)
 	}
     }
 
-  if ([wData length] > 0)
+  /* Ensure we set the correct content length (may be zero)
+   */
+  if ((id)NSMapGet(wProperties, (void*)@"Content-Length") == nil)
     {
       NSMapInsert(wProperties, (void*)@"Content-Length",
-	(void*)[NSString stringWithFormat: @"%d", [wData length]]);
+        (void*)[NSString stringWithFormat: @"%"PRIuPTR, [wData length]]);
+    }
 
+  if ([wData length] > 0)
+    {
       /*
        * Assume content type if not specified.
        */
@@ -459,6 +444,7 @@ static void debugWrite(GSHTTPURLHandle *handle, NSData *data)
 	    (void*)@"application/x-www-form-urlencoded");
 	}
     }
+
   if ((id)NSMapGet(wProperties, (void*)@"Authorization") == nil)
     {
       NSURLProtectionSpace	*space;
@@ -554,7 +540,7 @@ static void debugWrite(GSHTTPURLHandle *handle, NSData *data)
   /*
    * Send request to server.
    */
-  if (debug == YES) debugWrite(self, buf);
+  if (YES == debug) debugWrite(self, buf);
   [sock writeInBackgroundAndNotify: buf];
   RELEASE(buf);
   RELEASE(s);
@@ -574,7 +560,7 @@ static void debugWrite(GSHTTPURLHandle *handle, NSData *data)
   if (debug)
     NSLog(@"%@ %p %s", NSStringFromSelector(_cmd), self, keepalive?"K":"");
   d = [dict objectForKey: NSFileHandleNotificationDataItem];
-  if (debug == YES) debugRead(self, d);
+  if (YES == debug) debugRead(self, d);
   readCount = [d length];
 
   if (connectionState == idle)
@@ -584,7 +570,7 @@ static void debugWrite(GSHTTPURLHandle *handle, NSData *data)
        * it should just be the connection being closed by the other
        * end because of a timeout etc.
        */
-      if (debug == YES && [d length] != 0)
+      if (YES == debug && [d length] != 0)
 	{
 	  NSLog(@"%@ %p %s Unexpected data (%*.*s) from remote!",
 	    NSStringFromSelector(_cmd), self, keepalive?"K":"",
@@ -596,7 +582,7 @@ static void debugWrite(GSHTTPURLHandle *handle, NSData *data)
     }
   else if ([parser parse: d] == NO && [parser isComplete] == NO)
     {
-      if (debug == YES)
+      if (YES == debug)
 	{
 	  NSLog(@"HTTP parse failure - %@", parser);
 	}
@@ -765,6 +751,7 @@ static void debugWrite(GSHTTPURLHandle *handle, NSData *data)
 		    {
 		      [self writeProperty: auth forKey: @"Authorization"];
 		      [self _tryLoadInBackground: u];
+                      RELEASE(self);
 		      return;	// Retrying.
 		    }
 		}
@@ -826,7 +813,7 @@ static void debugWrite(GSHTTPURLHandle *handle, NSData *data)
 	   * lost in the network or the remote end received it and
 	   * the response was lost.
 	   */
-	  if (debug == YES)
+	  if (YES == debug)
 	    {
 	      NSLog(@"HTTP response not received - %@", parser);
 	    }
@@ -856,7 +843,7 @@ static void debugWrite(GSHTTPURLHandle *handle, NSData *data)
   if (debug)
     NSLog(@"%@ %p %s", NSStringFromSelector(_cmd), self, keepalive?"K":"");
   d = [dict objectForKey: NSFileHandleNotificationDataItem];
-  if (debug == YES) debugRead(self, d);
+  if (YES == debug) debugRead(self, d);
 
   if ([d length] > 0)
     {
@@ -1012,7 +999,7 @@ static void debugWrite(GSHTTPURLHandle *handle, NSData *data)
                object: sock];
 
       buf = [cmd dataUsingEncoding: NSASCIIStringEncoding];
-      if (debug == YES) debugWrite(self, buf);
+      if (YES == debug) debugWrite(self, buf);
       [sock writeInBackgroundAndNotify: buf];
 
       when = [NSDate alloc];
@@ -1041,16 +1028,52 @@ static void debugWrite(GSHTTPURLHandle *handle, NSData *data)
     }
   if ([[u scheme] isEqualToString: @"https"])
     {
+      static NSArray            *keys = nil;
+      NSMutableDictionary       *opts;
+      NSUInteger                count;
+
       /* If we are an https connection, negotiate secure connection.
        * Make sure we are not an observer of the file handle while
        * it is connecting...
        */
       [nc removeObserver: self name: nil object: sock];
+
+      if (nil == keys)
+        {
+          keys = [[NSArray alloc] initWithObjects:
+            GSTLSCAFile,
+            GSTLSCertificateFile,
+            GSTLSCertificateKeyFile,
+            GSTLSCertificateKeyPassword,
+            GSTLSDebug,
+            GSTLSPriority,
+            GSTLSRemoteHosts,
+            GSTLSRevokeFile,
+            GSTLSVerify,
+            nil];
+        }
+      count = [keys count];
+      opts = [[NSMutableDictionary alloc] initWithCapacity: count];
+      while (count-- > 0)
+        {
+          NSString      *key = [keys objectAtIndex: count];
+          NSString      *str = [request objectForKey: key];
+
+          if (nil != str)
+            {
+              [opts setObject: str forKey: key];
+            }
+        }
+      if (debug) [opts setObject: @"YES" forKey: GSTLSDebug];
+      [sock sslSetOptions: opts];
+      [opts release];
       if ([sock sslConnect] == NO)
 	{
 	  if (debug)
 	    NSLog(@"%@ %p %s Failed to make ssl connect",
 	      NSStringFromSelector(_cmd), self, keepalive?"K":"");
+          [sock closeFile];
+          DESTROY(sock);
 	  [self endLoadInBackground];
 	  [self backgroundLoadDidFailWithReason:
 	    @"Failed to make ssl connect"];
@@ -1407,6 +1430,8 @@ static void debugWrite(GSHTTPURLHandle *handle, NSData *data)
 	  if ([[u scheme] isEqualToString: @"https"])
 	    {
 	      NSString	*cert;
+              NSString	*key;
+              NSString	*pwd;
 
 	      if (sslClass == 0)
 		{
@@ -1417,16 +1442,27 @@ static void debugWrite(GSHTTPURLHandle *handle, NSData *data)
 	      sock = [sslClass fileHandleAsClientInBackgroundAtAddress: host
 							       service: port
 							      protocol: s];
-	      cert = [request objectForKey: GSHTTPPropertyCertificateFileKey];
-	      if ([cert length] > 0)
-		{
-		  NSString	*key;
-		  NSString	*pwd;
 
-		  key = [request objectForKey: GSHTTPPropertyKeyFileKey];
-		  pwd = [request objectForKey: GSHTTPPropertyPasswordKey];
-		  [sock sslSetCertificate: cert privateKey: key PEMpasswd: pwd];
-		}
+              /* Map old SSL keys onto new.
+               */
+	      cert = [request objectForKey: GSHTTPPropertyCertificateFileKey];
+	      if (nil != cert)
+		{
+                  [request setObject: cert
+                              forKey: GSTLSCertificateFile];
+                }
+              key = [request objectForKey: GSHTTPPropertyKeyFileKey];
+              if (nil != key)
+                {
+                  [request setObject: key
+                              forKey: GSTLSCertificateKeyFile];
+                }
+              pwd = [request objectForKey: GSHTTPPropertyPasswordKey];
+              if (nil != pwd)
+                {
+                  [request setObject: pwd
+                              forKey: GSTLSCertificateKeyPassword];
+                }
 	    }
 	  else
 	    {
