@@ -32,13 +32,13 @@
 #import "Foundation/NSRunLoop.h"
 #import "Foundation/NSTimer.h"
 #import "Foundation/NSValue.h"
-#import "GNUstepBase/NSObject+GNUstepBase.h"
 #if defined(_REENTRANT)
 #import "GNUstepBase/GSLock.h"
 #endif
 
 #import <dns_sd.h>		// Apple's DNS Service Discovery
 
+#import <sys/select.h>
 #import <sys/types.h>
 #import <sys/socket.h>		// AF_INET / AF_INET6
 
@@ -793,7 +793,7 @@ static void DNSSD_API
       Browser	*browser;
       
       browser = malloc(sizeof (struct _Browser));
-      memset(browser, 0, sizeof browser);
+      memset(browser, 0, sizeof &browser);
       
       CREATELOCK(browser);
       
@@ -1735,7 +1735,7 @@ static void DNSSD_API
       Service	*service;
       
       service = malloc(sizeof (struct _Service));
-      memset(service, 0, sizeof service);
+      memset(service, 0, sizeof &service);
       
       CREATELOCK(service);
       
@@ -1744,7 +1744,7 @@ static void DNSSD_API
       service->timer = nil;
       service->timeout = nil;
       
-      service->info = [[NSMutableDictionary alloc] initWithCapacity: 1];
+      service->info = [[NSMutableDictionary alloc] initWithCapacity: 3];
       [service->info setObject: [domain retain]
 			forKey: @"Domain"];
       [service->info setObject: [name retain]
@@ -1870,12 +1870,6 @@ static void DNSSD_API
 	    break;
 	  }
 	
-	if (! [self delegate])
-	  {
-	    err = NSNetServicesInvalidError;
-	    break;
-	  }
-	
 	if (service->timer)
 	  {
 	    err = NSNetServicesActivityInProgress;
@@ -2478,7 +2472,7 @@ static void DNSSD_API
       Monitor	*monitor;
       
       monitor = malloc(sizeof (struct _Monitor));
-      memset(monitor, 0, sizeof monitor);
+      memset(monitor, 0, sizeof &monitor);
       
       CREATELOCK(monitor);
       
