@@ -28,20 +28,7 @@
 #import	"Foundation/NSNetServices.h"
 #import "GNUstepBase/NSNetServices+GNUstepBase.h"
 
-// Subclasses using mDNSResponder:
-
-/**
- * NSNetService using the mDNSResponder API.
- */
-@interface GSMDNSNetService : NSNetService
-@end
-
-
-/**
- * NSNetServiceBrowser using the mDNSResponder API.
- */
-@interface GSMDNSNetServiceBrowser : NSNetServiceBrowser
-@end
+#if GS_USE_AVAHI==1
 
 // Subclasses using Avahi:
 
@@ -78,12 +65,20 @@ typedef enum
  */
 NSString* GSNetServiceDotTerminatedNSStringFromString(const char* string);
 
-@class GSAvahiRunLoopContext, NSTimer, NSLock, NSRecursiveLock, NSMutableDictionary, NSMapTable;
+@class GSAvahiRunLoopContext;
+@class NSTimer;
+@class NSRecursiveLock;
+@class NSMutableDictionary;
+@class NSMapTable;
 
 /**
  * NSNetService using the avahi-client API.
  */
+#ifdef __clang__        /* FIXME ... this is not clang specific! */
+@interface GSAvahiNetService : NSNetService <NSNetServiceDelegate>
+#else
 @interface GSAvahiNetService : NSNetService
+#endif
 {
   // GSAvahiClient behaviour ivars:
   // From superclass: id _delegate;
@@ -92,7 +87,7 @@ NSString* GSNetServiceDotTerminatedNSStringFromString(const char* string);
   NSRecursiveLock *_lock;
   // Ivars for this class:
   NSMutableDictionary *_info;
-  NSLock *_infoLock;
+  NSRecursiveLock *_infoLock;
   NSUInteger _infoSeq;
   GSNetServiceState _serviceState;
   int _ifIndex;
@@ -125,7 +120,12 @@ NSString* GSNetServiceDotTerminatedNSStringFromString(const char* string);
 /**
  * NSNetServiceBrowser using the avahi-client API.
  */
+#ifdef __clang__        /* FIXME ... this is not clang specific! */
+@interface GSAvahiNetServiceBrowser
+  : NSNetServiceBrowser <NSNetServiceBrowserDelegate>
+#else
 @interface GSAvahiNetServiceBrowser: NSNetServiceBrowser
+#endif
 {
   // GSAvahiClient behaviour ivars:
   // from superclass: id _delegate;
@@ -139,3 +139,29 @@ NSString* GSNetServiceDotTerminatedNSStringFromString(const char* string);
   NSMutableDictionary *_services;
 }
 @end
+
+#else // GS_USE_MDNS
+
+// Subclasses using mDNSResponder:
+
+/**
+ * NSNetService using the mDNSResponder API.
+ */
+#ifdef __clang__        /* FIXME ... this is not clang specific! */
+@interface GSMDNSNetService : NSNetService <NSNetServiceDelegate>
+#else
+@interface GSMDNSNetService : NSNetService
+#endif
+@end
+
+/**
+ * NSNetServiceBrowser using the mDNSResponder API.
+ */
+#ifdef __clang__        /* FIXME ... this is not clang specific! */
+@interface GSMDNSNetServiceBrowser : NSNetServiceBrowser <NSNetServiceBrowserDelegate>
+#else
+@interface GSMDNSNetServiceBrowser : NSNetServiceBrowser
+#endif
+@end
+
+#endif // GS_USE_AVAHI
