@@ -95,7 +95,7 @@ typedef struct objc_category* Category;
  * arrays are allocated on the stack (for speed), but large arrays are
  * allocated from the heap (to avoid stack overflow).
  */
-#if __GNUC__ > 3
+#if __GNUC__ > 3 && !defined(__clang__)
 __attribute__((unused)) static void GSFreeTempBuffer(void **b)
 {
   if (NULL != *b) free(*b);
@@ -110,8 +110,10 @@ __attribute__((unused)) static void GSFreeTempBuffer(void **b)
       P = _base;\
     }
 #else
+/* Make minimum size of _ibuf 1 to avoid compiler warnings.
+ */
 #  define	GS_BEGINITEMBUF(P, S, T) { \
-  T _ibuf[(S) <= GS_MAX_OBJECTS_FROM_STACK ? (S) : 0]; \
+  T _ibuf[(S) > 0 && (S) <= GS_MAX_OBJECTS_FROM_STACK ? (S) : 1]; \
   T *_base = ((S) <= GS_MAX_OBJECTS_FROM_STACK) ? _ibuf \
     : (T*)NSZoneMalloc(NSDefaultMallocZone(), (S) * sizeof(T)); \
   T *(P) = _base;
@@ -122,7 +124,7 @@ __attribute__((unused)) static void GSFreeTempBuffer(void **b)
  * arrays of items.  Use GS_BEGINITEMBUF() to start the block of code using
  * the array and this macro to end it.
  */
-#if __GNUC__ > 3
+#if __GNUC__ > 3 && !defined(__clang__)
 # define	GS_ENDITEMBUF() }
 #else
 #  define	GS_ENDITEMBUF() \
@@ -376,34 +378,6 @@ GSPrivateIsByteEncoding(NSStringEncoding encoding) GS_ATTRIB_PRIVATE;
 BOOL
 GSPrivateIsEncodingSupported(NSStringEncoding encoding) GS_ATTRIB_PRIVATE;
 
-/* Hash function to hash up to limit bytes from data of specified length.
- * If the flag is NO then a result of 0 is mapped to 0xffffffff.
- * This is a pretty useful general purpose hash function.
- */
-static inline unsigned
-GSPrivateHash(const void *data, unsigned length, unsigned limit, BOOL zero)
-  __attribute__((unused));
-static inline unsigned
-GSPrivateHash(const void *data, unsigned length, unsigned limit, BOOL zero)
-{
-  unsigned	ret = length;
-  unsigned	l = length;
-
-  if (limit < length)
-    {
-      l = limit;
-    }
-  while (l-- > 0)
-    {
-      ret = (ret << 5) + ret + ((const unsigned char*)data)[l];
-    }
-  if (ret == 0 && zero == NO)
-    {
-       ret = 0xffffffff;
-    }
-  return ret;
-}
-
 /* load a module into the runtime
  */
 long
@@ -527,12 +501,14 @@ GSPrivateUnloadModule(FILE *errorStream,
   unsigned      size;
   void          *buffer;
   void		*executable;
+  id            frame;
 }
 + (GSCodeBuffer*) memoryWithSize: (NSUInteger)_size;
 - (void*) buffer;
 - (void*) executable;
 - (id) initWithSize: (NSUInteger)_size;
 - (void) protect;
+- (void) setFrame: (id)aFrame;
 @end
 
 BOOL
@@ -541,5 +517,28 @@ GSPrivateIsCollectable(const void *ptr) GS_ATTRIB_PRIVATE;
 NSZone*
 GSAtomicMallocZone (void);
 
+/* Generate a 32bit hash from supplied byte data.
+ */
+uint32_t
+GSPrivateHash(uint32_t seed, const void *bytes, int length)
+  GS_ATTRIB_PRIVATE;
+
+/* Incorporate 'l' bytes of data from the buffer pointed to by 'b' into
+ * the hash state information pointed to by p0 and p1.
+ * The hash state variables should have been initialised to zero before
+ * the first call to this function, and the result should be produced
+ * by calling the GSPrivateFinishHash() function.
+ */
+void
+GSPrivateIncrementalHash(uint32_t *p0, uint32_t *p1, const void *b, int l)
+  GS_ATTRIB_PRIVATE;
+
+/* Generate a 32bit hash from supplied state variables resulting from
+ * calls to the GSPrivateIncrementalHash() function.
+ */
+uint32_t
+GSPrivateFinishHash(uint32_t s0, uint32_t s1, uint32_t totalLength)
+  GS_ATTRIB_PRIVATE;
+
 #endif /* _GSPrivate_h_ */
 
