@@ -28,7 +28,7 @@
 
 
    <title>NSBundle class reference</title>
-   $Date: 2012-01-16 01:15:07 -0700 (Mon, 16 Jan 2012) $ $Revision: 34555 $
+   $Date: 2013-08-22 20:55:03 +0100 (Thu, 22 Aug 2013) $ $Revision: 37008 $
 */
 
 #define	EXPOSE_NSBundle_IVARS	1
@@ -51,7 +51,6 @@
 #import "Foundation/NSData.h"
 #import "Foundation/NSURL.h"
 #import "Foundation/NSValue.h"
-#import "GNUstepBase/NSObject+GNUstepBase.h"
 #import "GNUstepBase/NSString+GNUstepBase.h"
 
 #import "GSPrivate.h"
@@ -61,8 +60,6 @@ NSString * const NSBundleDidLoadNotification = @"NSBundleDidLoadNotification";
 NSString * const NSShowNonLocalizedStrings = @"NSShowNonLocalizedStrings";
 NSString * const NSLoadedClasses = @"NSLoadedClasses";
 
-#include <string.h>
-
 static NSFileManager	*
 manager()
 {
@@ -71,10 +68,106 @@ manager()
   if (mgr == nil)
     {
       mgr = RETAIN([NSFileManager defaultManager]);
+      [[NSObject leakAt: &mgr] release];
     }
   return mgr;
 }
 
+static NSDictionary     *langAliases = nil;
+static NSDictionary     *langCanonical = nil;
+
+/* Map a language name to any alternative versions.   This function should
+ * return an array of alternative language/localisation directory names in
+ * the preferred order of precedence (ie resources in the directories named
+ * earlier in the array are to be preferred to those in directories named
+ * later).
+ * We should support regional language specifications (such as en-GB)
+ * as our first priority, and then fall back to the more general names.
+ * NB. Also handle the form  like en-GB_US (English language, British dialect,
+ * in the United States region).
+ */
+static NSArray *
+altLang(NSString *full)
+{
+  NSMutableArray        *a = nil;
+
+  if (nil != full)
+    {
+      NSString  *alias = nil;
+      NSString  *canon = nil;
+      NSString  *lang = nil;
+      NSString  *dialect = nil;
+      NSString  *region = nil;
+      NSRange   r;
+
+      alias = [langAliases objectForKey: full];
+      if (nil == alias)
+        {
+          canon = [langCanonical objectForKey: full];
+          if (nil != canon)
+            {
+              alias = [langAliases objectForKey: canon];
+            }
+          if (nil == alias)
+            {
+              alias = full;
+            }
+        }
+      canon = [langCanonical objectForKey: alias];
+      if (nil == canon)
+        {
+          canon = [langCanonical objectForKey: full];
+          if (nil == canon)
+            {
+              canon = full;
+            }
+        }
+
+      if ((r = [canon rangeOfString: @"-"]).length > 1)
+        {
+          dialect = [canon substringFromIndex: NSMaxRange(r)];
+          lang = [canon substringToIndex: r.location];
+          if ((r = [dialect rangeOfString: @"_"]).length > 1)
+            {
+              region = [dialect substringFromIndex: NSMaxRange(r)];
+              dialect = [dialect substringToIndex: r.location];
+            }
+        }
+      else if ((r = [canon rangeOfString: @"_"]).length > 1)
+        {
+          region = [canon substringFromIndex: NSMaxRange(r)];
+          lang = [canon substringToIndex: r.location];
+        }
+      else
+        {
+          lang = canon;
+        }
+
+      a = [NSMutableArray arrayWithCapacity: 5];
+      if (nil != dialect && nil != region)
+        {
+          [a addObject: [NSString stringWithFormat: @"%@-%@_%@",
+            lang, dialect, region]];
+        }
+      if (nil != dialect)
+        {
+          [a addObject: [NSString stringWithFormat: @"%@-%@",
+            lang, dialect]];
+        }
+      if (nil != region)
+        {
+          [a addObject: [NSString stringWithFormat: @"%@_%@",
+            lang, region]];
+        }
+      [a addObject: lang];
+      if (NO == [a containsObject: alias])
+        {
+          [a addObject: alias];
+        }
+    }
+  return a;
+}
+
 static NSLock *pathCacheLock = nil;
 static NSMutableDictionary *pathCache = nil;
 
@@ -95,6 +188,7 @@ typedef enum {
 /* Class variables - We keep track of all the bundles */
 static NSBundle		*_mainBundle = nil;
 static NSMapTable	*_bundles = NULL;
+static NSMapTable	*_byClass = NULL;
 static NSMapTable	*_byIdentifier = NULL;
 
 /* Store the working directory at startup */
@@ -399,8 +493,9 @@ addBundlePath(NSMutableArray *list, NSArray *contents,
     }
   if (nil != subdir)
     {
-      NSEnumerator *e = [[subdir pathComponents] objectEnumerator];
-      NSString *subdirComponent;
+      NSEnumerator      *e = [[subdir pathComponents] objectEnumerator];
+      NSString          *subdirComponent;
+
       while ((subdirComponent = [e nextObject]) != nil)
 	{
 	  if (NO == [contents containsObject: subdirComponent])
@@ -414,20 +509,30 @@ addBundlePath(NSMutableArray *list, NSArray *contents,
 	    }	  
 	}
     }
-  if (nil != lang)
+  if (nil == lang)
     {
-      lang = [lang stringByAppendingPathExtension: @"lproj"];
-      if (NO == [contents containsObject: lang])
-	{
-	  return;
-	}
-      path = [path stringByAppendingPathComponent: lang];
-      if (nil == (contents = bundle_directory_readable(path)))
-	{
-	  return;
-	}
+      [list addObject: path];
+    }
+  else
+    {
+      NSEnumerator      *enumerator = [altLang(lang) objectEnumerator];
+      NSString          *alt;
+
+      /* Add each language specific subdirectory in order.
+       */
+      while (nil != (alt = [enumerator nextObject]))
+        {
+          alt = [alt stringByAppendingPathExtension: @"lproj"];
+          if (YES == [contents containsObject: alt])
+            {
+              alt = [path stringByAppendingPathComponent: alt];
+              if (nil != (contents = bundle_directory_readable(alt)))
+                {
+                  [list addObject: alt];
+                }
+            }
+        }
     }
-  [list addObject: path];
 }
 
 /* Try to locate name framework in standard places
@@ -517,7 +622,8 @@ _find_main_bundle_for_tool(NSString *toolName)
 
 @interface NSBundle (Private)
 + (NSString *) _absolutePathOfExecutable: (NSString *)path;
-+ (void) _addFrameworkFromClass: (Class)frameworkClass;
++ (NSBundle*) _addFrameworkFromClass: (Class)frameworkClass;
++ (NSMutableArray*) _addFrameworks;
 + (NSString*) _gnustep_target_cpu;
 + (NSString*) _gnustep_target_dir;
 + (NSString*) _gnustep_target_os;
@@ -577,17 +683,17 @@ _find_main_bundle_for_tool(NSString *toolName)
    disk to those framework bundles.
 
 */
-+ (void) _addFrameworkFromClass: (Class)frameworkClass
++ (NSBundle*) _addFrameworkFromClass: (Class)frameworkClass
 {
   NSBundle	*bundle = nil;
   NSString	**fmClasses;
   NSString	*bundlePath = nil;
   unsigned int	len;
-  const char *frameworkClassName;
+  const char    *frameworkClassName;
 
   if (frameworkClass == Nil)
     {
-      return;
+      return nil;
     }
 
   frameworkClassName = class_getName(frameworkClass);
@@ -595,11 +701,25 @@ _find_main_bundle_for_tool(NSString *toolName)
   len = strlen (frameworkClassName);
 
   if (len > 12 * sizeof(char)
-      && !strncmp ("NSFramework_", frameworkClassName, 12))
+    && !strncmp ("NSFramework_", frameworkClassName, 12))
     {
       /* The name of the framework.  */
       NSString *name;
 
+      /* If the bundle for this framework class is already loaded,
+       * simply return it.  The lookup will return an NSNull object
+       * if the framework class has no known bundle.
+       */
+      bundle = (id)NSMapGet(_byClass, frameworkClass);
+      if (nil != bundle)
+        {
+          if ((id)bundle == (id)[NSNull null])
+            {
+              bundle = nil;
+            }
+          return bundle;
+        }
+
       name = [NSString stringWithUTF8String: &frameworkClassName[12]];
       /* Important - gnustep-make mangles framework names to encode
        * them as ObjC class names.  Here we need to demangle them.  We
@@ -618,7 +738,7 @@ _find_main_bundle_for_tool(NSString *toolName)
 
       if ([bundlePath isEqualToString: GSPrivateExecutablePath()])
 	{
-	  /* Ops ... the NSFramework_xxx class is linked in the main
+	  /* Oops ... the NSFramework_xxx class is linked in the main
 	   * executable.  Maybe the framework was statically linked
 	   * into the application ... resort to searching the
 	   * framework bundle on the filesystem manually.
@@ -751,12 +871,21 @@ _find_main_bundle_for_tool(NSString *toolName)
 	    }
 	}
 
+      [load_lock lock];
       if (bundle == nil)
 	{
+          NSMapInsert(_byClass, frameworkClass, [NSNull null]);
+          [load_lock unlock];
 	  NSWarnMLog (@"Could not find framework %@ in any standard location",
 	    name);
-	  return;
+	  return nil;
 	}
+      else
+        {
+          bundle->_principalClass = frameworkClass;
+          NSMapInsert(_byClass, frameworkClass, bundle);
+          [load_lock unlock];
+        }
 
       bundle->_bundleType = NSBUNDLE_FRAMEWORK;
       bundle->_codeLoaded = YES;
@@ -773,6 +902,7 @@ _find_main_bundle_for_tool(NSString *toolName)
 	  NSValue *value;
 	  Class    class = NSClassFromString(*fmClasses);
 
+          NSMapInsert(_byClass, class, bundle);
 	  value = [NSValue valueWithPointer: (void*)class];
 	  [bundle->_bundleClasses addObject: value];
 
@@ -816,6 +946,39 @@ _find_main_bundle_for_tool(NSString *toolName)
 	    }
 	}
     }
+  return bundle;
+}
+
++ (NSMutableArray*) _addFrameworks
+{
+  int                   i;
+  int                   numClasses = 0;
+  int                   newNumClasses;
+  Class                 *classes = NULL;
+  NSMutableArray        *added = nil;
+
+  newNumClasses = objc_getClassList(NULL, 0);
+  while (numClasses < newNumClasses)
+    {
+      numClasses = newNumClasses;
+      classes = realloc(classes, sizeof(Class) * numClasses);
+      newNumClasses = objc_getClassList(classes, numClasses);
+    }
+  for (i = 0; i < numClasses; i++)
+    {
+      NSBundle  *bundle = [self _addFrameworkFromClass: classes[i]];
+
+      if (nil != bundle)
+        {
+          if (nil == added)
+            {
+              added = [NSMutableArray arrayWithCapacity: 100];
+            }
+          [added addObject: bundle];
+        }
+    }
+  free(classes);
+  return added;
 }
 
 + (NSString*) _gnustep_target_cpu
@@ -865,8 +1028,8 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
 
   /* Don't store the internal NSFramework_xxx class into the list of
      bundle classes, but store the linked frameworks in _loadingFrameworks  */
-  if (strlen (className) > 12   &&  !strncmp ("NSFramework_",
-						   className, 12))
+  if (strlen (className) > 12
+    &&  !strncmp ("NSFramework_", className, 12))
     {
       if (_currentFrameworkName)
 	{
@@ -891,11 +1054,36 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
 
 @implementation NSBundle
 
++ (void) atExit
+{
+  if ([NSObject shouldCleanUp])
+    {
+      DESTROY(_emptyTable);
+      DESTROY(langAliases);
+      DESTROY(langCanonical);
+      DESTROY(_bundles);
+      DESTROY(_byClass);
+      DESTROY(_byIdentifier);
+      DESTROY(pathCache);
+      DESTROY(pathCacheLock);
+      DESTROY(load_lock);
+      DESTROY(gnustep_target_cpu);
+      DESTROY(gnustep_target_os);
+      DESTROY(gnustep_target_dir);
+      DESTROY(library_combo);
+      DESTROY(_launchDirectory);
+      DESTROY(_gnustep_bundle);
+      DESTROY(_mainBundle);
+    }
+}
+
 + (void) initialize
 {
   if (self == [NSBundle class])
     {
       extern const char	*GSPathHandling(const char *);
+      NSAutoreleasePool *pool = [NSAutoreleasePool new];
+      NSString          *file;
       const char	*mode;
       NSDictionary	*env;
       NSString		*str;
@@ -903,12 +1091,75 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
       /* Ensure we do 'right' path handling while initializing core paths.
        */
       mode = GSPathHandling("right");
-      _emptyTable = RETAIN([NSDictionary dictionary]);
+      _emptyTable = [NSDictionary new];
+
+      /* Create basic mapping dictionaries for bootstrapping and
+       * for use if the full ductionaries can't be loaded from the
+       * gnustep-base library resource bundle.
+       */
+      langAliases = [[NSDictionary alloc] initWithObjectsAndKeys:
+        @"Dutch", @"nl",
+        @"English", @"en",
+        @"Esperanto", @"eo",
+        @"French", @"fr",
+        @"German", @"de",
+        @"Hungarian", @"hu",
+        @"Italian", @"it",
+        @"Korean", @"ko",
+        @"Russian", @"ru",
+        @"Slovak", @"sk",
+        @"Spanish", @"es",
+        @"TraditionalChinese", @"zh",
+        @"Ukrainian", @"uk",
+        nil];
+      langCanonical = [[NSDictionary alloc] initWithObjectsAndKeys:
+        @"de", @"German",
+        @"de", @"ger",
+        @"de", @"deu",
+        @"en", @"English",
+        @"en", @"eng",
+        @"ep", @"Esperanto",
+        @"ep", @"epo",
+        @"ep", @"epo",
+        @"fr", @"French",
+        @"fr", @"fra",
+        @"fr", @"fre",
+        @"hu", @"Hungarian",
+        @"hu", @"hun",
+        @"it", @"Italian",
+        @"it", @"ita",
+        @"ko", @"Korean",
+        @"ko", @"kir",
+        @"nl", @"Dutch",
+        @"nl", @"dut",
+        @"nl", @"nld",
+        @"ru", @"Russian",
+        @"ru", @"rus",
+        @"sk", @"Slovak",
+        @"sk", @"slo",
+        @"sk", @"slk",
+        @"sp", @"Spanish",
+        @"sp", @"spa",
+        @"uk", @"Ukrainian",
+        @"uk", @"ukr",
+        @"zh", @"TraditionalChinese",
+        @"zh", @"chi",
+        @"zh", @"zho",
+        nil];
 
       /* Initialise manager here so it's thread-safe.
        */
       manager();
 
+      /* Set up tables for bundle lookups
+       */
+      _bundles = NSCreateMapTable(NSObjectMapKeyCallBacks,
+	NSNonOwnedPointerMapValueCallBacks, 0);
+      _byClass = NSCreateMapTable(NSNonOwnedPointerMapKeyCallBacks,
+	NSNonOwnedPointerMapValueCallBacks, 0);
+      _byIdentifier = NSCreateMapTable(NSObjectMapKeyCallBacks,
+	NSNonOwnedPointerMapValueCallBacks, 0);
+
       pathCacheLock = [NSLock new];
       pathCache = [NSMutableDictionary new];
 
@@ -947,45 +1198,49 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
 
       _gnustep_bundle = RETAIN([self bundleForLibrary: @"gnustep-base"
 					      version: _base_version]);
+
+      /* The Locale aliases map converts canonical names to old-style names
+       */
+      file = [_gnustep_bundle pathForResource: @"Locale"
+                                       ofType: @"aliases"
+                                  inDirectory: @"Languages"];
+      if (file != nil)
+        {
+          NSDictionary  *d;
+
+          d = [[NSDictionary alloc] initWithContentsOfFile: file];
+          if ([d count] > 0)
+            {
+              ASSIGN(langAliases, d);
+            }
+          [d release];
+        }
+
+      /* The Locale canonical map converts old-style names to ISO 639 names
+       * and converts ISO 639-2 names to the preferred ISO 639-1 names where
+       * an ISO 639-1 name exists.
+       */
+      file = [_gnustep_bundle pathForResource: @"Locale"
+                                       ofType: @"canonical"
+                                  inDirectory: @"Languages"];
+      if (file != nil)
+        {
+          NSDictionary  *d;
+
+          d = [[NSDictionary alloc] initWithContentsOfFile: file];
+          if ([d count] > 0)
+            {
+              ASSIGN(langCanonical, d);
+            }
+          [d release];
+        }
+
 #if 0
       _loadingBundle = [self mainBundle];
       handle = objc_open_main_module(stderr);
       printf("%08x\n", handle);
 #endif
-#if NeXT_RUNTIME || defined(__GNUSTEP_RUNTIME__) || defined(__GNU_LIBOBJC__)
-      {
-	int i, numClasses = 0, newNumClasses = objc_getClassList(NULL, 0);
-	Class *classes = NULL;
-	while (numClasses < newNumClasses)
-	  {
-	    numClasses = newNumClasses;
-	    classes = realloc(classes, sizeof(Class) * numClasses);
-	    newNumClasses = objc_getClassList(classes, numClasses);
-	  }
-	for (i = 0; i < numClasses; i++)
-	  {
-	    [self _addFrameworkFromClass: classes[i]];
-	  }
-	free(classes);
-      }
-#else
-      {
-	void	*state = NULL;
-	Class	class;
-
-	while ((class = objc_next_class(&state)))
-	  {
-	    const char *className = class_getName(class);
-	    unsigned int len = strlen (className);
-
-	    if (len > sizeof("NSFramework_")
-		&& !strncmp("NSFramework_", className, 12))
-	      {
-		[self _addFrameworkFromClass: class];
-	      }
-	  }
-      }
-#endif
+      [self _addFrameworks];
 #if 0
       //  _bundle_load_callback(class, NULL);
       //  bundle = (NSBundle *)NSMapGet(_bundles, bundlePath);
@@ -994,6 +1249,8 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
       _loadingBundle = nil;
 #endif
       GSPathHandling(mode);
+      [pool release];
+      [self registerAtExit];
     }
 }
 
@@ -1004,32 +1261,30 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
 + (NSArray *) allBundles
 {
   NSMutableArray	*array = [NSMutableArray arrayWithCapacity: 2];
+  NSMapEnumerator	enumerate;
+  void		        *key;
+  NSBundle		*bundle;
 
   [load_lock lock];
   if (!_mainBundle)
     {
       [self mainBundle];
     }
-  if (_bundles != 0)
-    {
-      NSMapEnumerator	enumerate;
-      void		*key;
-      NSBundle		*bundle;
 
-      enumerate = NSEnumerateMapTable(_bundles);
-      while (NSNextMapEnumeratorPair(&enumerate, &key, (void **)&bundle))
-	{
-	  if (bundle->_bundleType == NSBUNDLE_FRAMEWORK)
-	    {
-	      continue;
-	    }
-	  if ([array indexOfObjectIdenticalTo: bundle] == NSNotFound)
-	    {
-	      [array addObject: bundle];
-	    }
-	}
-      NSEndMapTableEnumeration(&enumerate);
+  enumerate = NSEnumerateMapTable(_bundles);
+  while (NSNextMapEnumeratorPair(&enumerate, &key, (void **)&bundle))
+    {
+      if (bundle->_bundleType == NSBUNDLE_FRAMEWORK)
+        {
+          continue;
+        }
+      if ([array indexOfObjectIdenticalTo: bundle] == NSNotFound)
+        {
+          [array addObject: bundle];
+        }
     }
+  NSEndMapTableEnumeration(&enumerate);
+
   [load_lock unlock];
   return array;
 }
@@ -1274,28 +1529,44 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
     }
 
   [load_lock lock];
-  bundle = nil;
-  enumerate = NSEnumerateMapTable(_bundles);
-  while (NSNextMapEnumeratorPair(&enumerate, &key, (void **)&bundle))
+  /* Try lookup ... if not found, make sure that all loaded bundles have
+   * class->bundle mapp entries set up and check again.
+   */
+  bundle = (NSBundle *)NSMapGet(_byClass, aClass);
+  if ((id)bundle == (id)[NSNull null])
     {
-      int i, j;
-      NSArray *bundleClasses = bundle->_bundleClasses;
-      BOOL found = NO;
+      [load_lock unlock];
+      return nil;
+    }
+  if (nil == bundle)
+    {
+      enumerate = NSEnumerateMapTable(_bundles);
+      while (NSNextMapEnumeratorPair(&enumerate, &key, (void **)&bundle))
+        {
+          NSArray           *classes = bundle->_bundleClasses;
+          NSUInteger        count = [classes count];
 
-      j = [bundleClasses count];
-      for (i = 0; i < j && found == NO; i++)
-	{
-	  if ([[bundleClasses objectAtIndex: i] pointerValue] == (void*)aClass)
-	    found = YES;
-	}
-
-      if (found == YES)
-	break;
-
-      bundle = nil;
+          if (count > 0
+            && 0 == NSMapGet(_byClass, [[classes lastObject] pointerValue]))
+            {
+              while (count-- > 0)
+                {
+                  NSMapInsert(_byClass,
+                    (void*)[[classes objectAtIndex: count] pointerValue],
+                    (void*)bundle);
+                }
+            }
+        }
+      NSEndMapTableEnumeration(&enumerate);
+      bundle = (NSBundle *)NSMapGet(_byClass, aClass);
+      if ((id)bundle == (id)[NSNull null])
+        {
+          [load_lock unlock];
+          return nil;
+        }
     }
 
-  if (bundle == nil)
+  if (nil == bundle)
     {
       /* Is it in the main bundle or a library? */
       if (!class_isMetaClass(aClass))
@@ -1315,11 +1586,21 @@ _bundle_load_callback(Class theClass, struct objc_category *theCategory)
 
 	  /*
 	   * Get the library bundle ... if there wasn't one then we
-	   * will assume the class was in the program executable and
-	   * return the mainBundle instead.
+	   * will check to see if it's in a newly loaded framework
+           * and if not, assume the class was in the program executable
+	   * and return the mainBundle instead.
 	   */
 	  bundle = [NSBundle bundleForLibrary: lib];
-	  if (bundle == nil)
+          if (nil == bundle && [[self _addFrameworks] count] > 0)
+            {
+              bundle = (NSBundle *)NSMapGet(_byClass, aClass);
+              if ((id)bundle == (id)[NSNull null])
+                {
+                  [load_lock unlock];
+                  return nil;
+                }
+            }
+	  if (nil == bundle)
 	    {
 	      bundle = [self mainBundle];
 	    }
@@ -1383,6 +1664,7 @@ IF_NO_GC(
 - (id) initWithPath: (NSString*)path
 {
   NSString	*identifier;
+  NSBundle	*bundle;
 
   self = [super init];
 
@@ -1450,17 +1732,13 @@ IF_NO_GC(
 
   /* check if we were already initialized for this directory */
   [load_lock lock];
-  if (_bundles)
+  bundle = (NSBundle *)NSMapGet(_bundles, path);
+  if (bundle != nil)
     {
-      NSBundle	*bundle = (NSBundle *)NSMapGet(_bundles, path);
-
-      if (bundle != nil)
-	{
-	  IF_NO_GC([bundle retain];)
-	  [load_lock unlock];
-	  [self dealloc];
-	  return bundle;
-	}
+      IF_NO_GC([bundle retain];)
+      [load_lock unlock];
+      [self dealloc];
+      return bundle;
     }
   [load_lock unlock];
 
@@ -1475,7 +1753,14 @@ IF_NO_GC(
 	}
     }
 
+  /* OK ... this is a new bundle ... need to insert it in the global map
+   * to be found by this path so that a leter call to -bundleIdentifier
+   * can work.
+   */
   _path = [path copy];
+  [load_lock lock];
+  NSMapInsert(_bundles, _path, self);
+  [load_lock unlock];
 
   if ([[[_path lastPathComponent] pathExtension] isEqual: @"framework"] == YES)
     {
@@ -1492,30 +1777,22 @@ IF_NO_GC(
   identifier = [self bundleIdentifier];
 
   [load_lock lock];
-  if (!_bundles)
-    {
-      _bundles = NSCreateMapTable(NSObjectMapKeyCallBacks,
-	NSNonOwnedPointerMapValueCallBacks, 0);
-    }
-  if (!_byIdentifier)
-    {
-      _byIdentifier = NSCreateMapTable(NSObjectMapKeyCallBacks,
-	NSNonOwnedPointerMapValueCallBacks, 0);
-    }
   if (identifier != nil)
     {
       NSBundle	*bundle = (NSBundle *)NSMapGet(_byIdentifier, identifier);
 
-      if (bundle != nil)
-	{
-	  IF_NO_GC([bundle retain];)
-	  [load_lock unlock];
-	  [self dealloc];
-	  return bundle;
-	}
-      NSMapInsert(_byIdentifier, identifier, self);
+      if (bundle != self)
+        {
+          if (bundle != nil)
+            {
+              IF_NO_GC([bundle retain];)
+              [load_lock unlock];
+              [self dealloc];
+              return bundle;
+            }
+          NSMapInsert(_byIdentifier, identifier, self);
+        }
     }
-  NSMapInsert(_bundles, _path, self);
   [load_lock unlock];
 
   return self;
@@ -1544,16 +1821,33 @@ IF_NO_GC(
   if (_path != nil)
     {
       NSString		*identifier = [self bundleIdentifier];
+      NSUInteger        count;
       NSUInteger	plen = [_path length];
       NSEnumerator	*enumerator;
       NSString		*path;
 
       [load_lock lock];
-      NSMapRemove(_bundles, _path);
+      if (_bundles != nil)
+        {
+          NSMapRemove(_bundles, _path);
+        }
       if (identifier != nil)
         {
 	  NSMapRemove(_byIdentifier, identifier);
         }
+      if (_principalClass != nil)
+        {
+	  NSMapRemove(_byClass, _principalClass);
+        }
+      if (_byClass != nil)
+        {
+          count = [_bundleClasses count];
+          while (count-- > 0)
+            {
+              NSMapRemove(_byClass,
+                [[_bundleClasses objectAtIndex: count] pointerValue]);
+            }
+        }
       [load_lock unlock];
 
       /* Clean up path cache for this bundle.
@@ -1610,6 +1904,11 @@ IF_NO_GC(
   return _path;
 }
 
+- (NSURL*) bundleURL
+{
+  return [NSURL fileURLWithPath: [self bundlePath]];
+}
+
 - (Class) classNamed: (NSString *)className
 {
   int     i, j;
@@ -1722,18 +2021,26 @@ IF_NO_GC(
       NSEnumerator   *classEnumerator;
       NSMutableArray *classNames;
       NSValue        *class;
+      NSBundle       *savedLoadingBundle;
 
+      /* Get the binary and set up fraework name if it is a framework.
+       */
       object = [self executablePath];
       if (object == nil || [object length] == 0)
 	{
 	  [load_lock unlock];
 	  return NO;
 	}
+      savedLoadingBundle = _loadingBundle;
       _loadingBundle = self;
       _bundleClasses = [[NSMutableArray alloc] initWithCapacity: 2];
-      _loadingFrameworks = RETAIN([NSMutableArray arrayWithCapacity: 2]);
 
-      /* This code is executed twice if a class linked in the bundle call a
+      if (nil == savedLoadingBundle)
+        {
+          _loadingFrameworks = [[NSMutableArray alloc] initWithCapacity: 2];
+        }
+
+      /* This code is executed twice if a class linked in the bundle calls a
 	 NSBundle method inside +load (-principalClass). To avoid this we set
 	 _codeLoaded before loading the bundle. */
       _codeLoaded = YES;
@@ -1741,8 +2048,12 @@ IF_NO_GC(
       if (GSPrivateLoadModule(object, stderr, _bundle_load_callback, 0, 0))
 	{
 	  _codeLoaded = NO;
-	  DESTROY(_loadingFrameworks);
-	  DESTROY(_currentFrameworkName);
+          _loadingBundle = savedLoadingBundle;
+          if (nil == _loadingBundle)
+            {
+              DESTROY(_loadingFrameworks);
+              DESTROY(_currentFrameworkName);
+            }
 	  [load_lock unlock];
 	  return NO;
 	}
@@ -1762,19 +2073,22 @@ IF_NO_GC(
 	 normally want all loaded bundles to appear when they call
 	 +allBundles.  */
       IF_NO_GC([self retain];)
-      _loadingBundle = nil;
-
-      DESTROY(_loadingFrameworks);
-      DESTROY(_currentFrameworkName);
 
       classNames = [NSMutableArray arrayWithCapacity: [_bundleClasses count]];
       classEnumerator = [_bundleClasses objectEnumerator];
       while ((class = [classEnumerator nextObject]) != nil)
 	{
+          NSMapInsert(_byClass, class, self);
 	  [classNames addObject:
 	    NSStringFromClass((Class)[class pointerValue])];
 	}
 
+      _loadingBundle = savedLoadingBundle;
+      if (nil == _loadingBundle)
+        {
+          DESTROY(_loadingFrameworks);
+          DESTROY(_currentFrameworkName);
+        }
       [load_lock unlock];
 
       [[NSNotificationCenter defaultCenter]
@@ -1798,7 +2112,6 @@ IF_NO_GC(
   if (NSDecrementExtraRefCountWasZero(self))
     {
       [self dealloc];
-      self = nil;
     }
   [load_lock unlock];
 }
@@ -1882,7 +2195,6 @@ IF_NO_GC(
     }
   if ([extension length] == 0)
     {
-      extension = nil;
       file = name;
     }
   else
@@ -2077,16 +2389,22 @@ IF_NO_GC(
     {
       /* Add all non-localized paths, plus ones in the particular localization
 	 (if there is one). */
-      NSString *theDir = [path stringByDeletingLastPathComponent];
+      NSString  *theDir = [path stringByDeletingLastPathComponent];
+      NSString  *last = [theDir lastPathComponent];
 
-      if ([[theDir pathExtension] isEqual: @"lproj"] == NO)
+      if ([[last pathExtension] isEqual: @"lproj"] == NO)
 	{
 	  [result addObject: path];
 	}
-      else if ([localizationName length] > 0
-	&& [[theDir lastPathComponent] hasPrefix: localizationName])
-	{
-	  [result insertObject: path atIndex: 0];
+      else
+        {
+          NSString      *lang = [last stringByDeletingPathExtension];
+          NSArray       *alternatives = altLang(lang);
+
+          if ([alternatives count] > 0)
+            {
+              [result addObject: path];
+            }
 	}
     }
 
@@ -2449,6 +2767,40 @@ IF_NO_GC(
   return object;
 }
 
+- (NSURL *) executableURL
+{
+  return [NSURL fileURLWithPath: [self executablePath]];
+}
+
+- (NSString *) pathForAuxiliaryExecutable: (NSString *) executableName
+{
+  NSString  *version = _frameworkVersion;
+
+  if (!version)
+    version = @"Current";
+
+  if (_bundleType == NSBUNDLE_FRAMEWORK)
+    {
+#if !defined(__MINGW__)
+      return [_path stringByAppendingPathComponent:
+                      [NSString stringWithFormat:@"Versions/%@/%@",
+                      version, executableName]];
+#else
+      return [_path stringByAppendingPathComponent: executableName];
+#endif
+    }
+  else
+    {
+      return [_path stringByAppendingPathComponent: executableName];
+    }
+}
+
+- (NSURL *) URLForAuxiliaryExecutable: (NSString *) executableName
+{
+  return [NSURL fileURLWithPath: [self pathForAuxiliaryExecutable:
+                       executableName]];
+}
+
 - (NSString *) resourcePath
 {
   NSString *version = _frameworkVersion;
@@ -2523,6 +2875,40 @@ IF_NO_GC(
     }
 }
 
+- (NSURL *) builtInPlugInsURL
+{
+  return [NSURL fileURLWithPath: [self builtInPlugInsPath]];
+}
+
+- (NSString *) privateFrameworksPath
+{
+  NSString  *version = _frameworkVersion;
+
+  if (!version)
+    version = @"Current";
+
+  if (_bundleType == NSBUNDLE_FRAMEWORK)
+    {
+#if !defined(__MINGW__)
+      return [_path stringByAppendingPathComponent:
+                      [NSString stringWithFormat:@"Versions/%@/PrivateFrameworks",
+                      version]];
+#else
+      return [_path stringByAppendingPathComponent: @"PrivateFrameworks"];
+#endif
+    }
+  else
+    {
+      return [_path stringByAppendingPathComponent: @"PrivateFrameworks"];
+    }
+}
+
+- (NSURL *) privateFrameworksURL
+{
+  return [NSURL fileURLWithPath: [self privateFrameworksPath]];
+}
+
+
 - (NSString*) bundleIdentifier
 {
   return [[self infoDictionary] objectForKey: @"CFBundleIdentifier"];
