@@ -24,7 +24,7 @@
    Boston, MA 02111 USA.
 
    <title>NSData class reference</title>
-   $Date: 2012-01-09 01:28:27 -0700 (Mon, 09 Jan 2012) $ $Revision: 34468 $
+   $Date: 2013-08-22 16:44:54 +0100 (Thu, 22 Aug 2013) $ $Revision: 37003 $
    */
 
 /* NOTES	-	Richard Frith-Macdonald 1997
@@ -84,7 +84,6 @@
 #import "Foundation/NSURL.h"
 #import "Foundation/NSValue.h"
 #import "GSPrivate.h"
-#import "GNUstepBase/NSObject+GNUstepBase.h"
 #include <stdio.h>
 
 #ifdef	HAVE_MMAP
@@ -141,7 +140,7 @@ static SEL	appendSel;
 static IMP	appendImp;
 
 static BOOL
-readContentsOfFile(NSString* path, void** buf, unsigned int* len, NSZone* zone)
+readContentsOfFile(NSString* path, void** buf, long* len, NSZone* zone)
 {
 #if defined(__MINGW__)
   const unichar	*thePath = 0;
@@ -176,6 +175,10 @@ readContentsOfFile(NSString* path, void** buf, unsigned int* len, NSZone* zone)
       goto failure;
     }
 
+  // FIXME: since fseek returns a long, this code will fail on files
+  // larger than ~2GB on 32-bit systems. Probably this entire function should
+  // be removed and NSFileHandle used instead. -Eric
+
   /*
    *	Seek to the end of the file.
    */
@@ -629,7 +632,7 @@ failure:
 - (id) initWithContentsOfFile: (NSString*)path
 {
   void		*fileBytes;
-  unsigned	fileLength;
+  long          fileLength;
 
 #if	GS_WITH_GC
   if (readContentsOfFile(path, &fileBytes, &fileLength, 0) == NO)
@@ -754,10 +757,20 @@ failure:
  */
 - (void) getBytes: (void*)buffer range: (NSRange)aRange
 {
-  unsigned	size = [self length];
+  NSUInteger	size = [self length];
 
   GS_RANGE_CHECK(aRange, size);
-  memcpy(buffer, [self bytes] + aRange.location, aRange.length);
+  if (aRange.length > 0)
+    {
+      const void	*bytes = [self bytes];
+
+      if (0 == bytes)
+	{
+	  [NSException raise: NSInternalInconsistencyException
+		      format: @"missing bytes in getBytes:range:"];
+	}
+      memcpy(buffer, bytes + aRange.location, aRange.length);
+    }
 }
 
 - (id) replacementObjectForPortCoder: (NSPortCoder*)aCoder
@@ -774,7 +787,7 @@ failure:
 - (NSData*) subdataWithRange: (NSRange)aRange
 {
   void		*buffer;
-  unsigned	l = [self length];
+  NSUInteger	l = [self length];
 
   GS_RANGE_CHECK(aRange, l);
 
@@ -796,8 +809,8 @@ failure:
 - (NSUInteger) hash
 {
   unsigned char	buf[64];
-  unsigned	l = [self length];
-  unsigned	ret =0;
+  NSUInteger	l = [self length];
+  NSUInteger	ret =0;
 
   l = MIN(l,64);
 
@@ -1866,7 +1879,7 @@ failure:
 
 - (void) encodeWithCoder: (NSCoder*)aCoder
 {
-  unsigned	length = [self length];
+  NSUInteger	length = [self length];
   void		*bytes = [self mutableBytes];
 
   if ([aCoder allowsKeyedCoding])
@@ -1877,7 +1890,7 @@ failure:
     }
   else
     {  
-      [aCoder encodeValueOfObjCType: @encode(unsigned int)
+      [aCoder encodeValueOfObjCType: @encode(NSUInteger)
 			         at: &length];
       if (length)
         {
@@ -1913,9 +1926,9 @@ failure:
     }
   else
     {  
-      unsigned  l;
+      NSUInteger  l;
 
-      [aCoder decodeValueOfObjCType: @encode(unsigned int) at: &l];
+      [aCoder decodeValueOfObjCType: @encode(NSUInteger) at: &l];
       if (l)
         {
           void *b;
@@ -1928,7 +1941,7 @@ failure:
           if (b == 0)
             {
               NSLog(@"[NSDataMalloc -initWithCoder:] unable to get %u bytes",
-                l);
+                (unsigned int)l);
               DESTROY(self);
               return nil;
             }
@@ -2012,11 +2025,16 @@ failure:
 - (void) appendBytes: (const void*)aBuffer
 	      length: (NSUInteger)bufferSize
 {
-  unsigned	oldLength = [self length];
-  void*		buffer;
+  NSUInteger	oldLength = [self length];
+  void		*buffer;
 
   [self setLength: oldLength + bufferSize];
   buffer = [self mutableBytes];
+  if (0 == buffer)
+    {
+      [NSException raise: NSInternalInconsistencyException
+		  format: @"missing bytes in appendBytes:length:"];
+    }
   memcpy(buffer + oldLength, aBuffer, bufferSize);
 }
 
@@ -2042,8 +2060,8 @@ failure:
 - (void) replaceBytesInRange: (NSRange)aRange
 		   withBytes: (const void*)bytes
 {
-  unsigned	size = [self length];
-  unsigned	need = NSMaxRange(aRange);
+  NSUInteger	size = [self length];
+  NSUInteger	need = NSMaxRange(aRange);
 
   if (aRange.location > size)
     {
@@ -2052,11 +2070,18 @@ failure:
     }
   if (aRange.length > 0)
     {
+      void	*buf = [self mutableBytes];
+
+      if (0 == buf)
+	{
+	  [NSException raise: NSInternalInconsistencyException
+	    format: @"missing bytes in replaceBytesInRange:withBytes:"];
+	}
       if (need > size)
 	{
 	  [self setLength: need];
 	}
-      memmove([self mutableBytes] + aRange.location, bytes, aRange.length);
+      memmove(buf + aRange.location, bytes, aRange.length);
     }
 }
 
@@ -2069,10 +2094,10 @@ failure:
 		   withBytes: (const void*)bytes
 		      length: (NSUInteger)length
 {
-  unsigned	size = [self length];
-  unsigned	end = NSMaxRange(aRange);
-  int		shift = length - aRange.length;
-  unsigned	need = size + shift;
+  NSUInteger	size = [self length];
+  NSUInteger	end = NSMaxRange(aRange);
+  NSInteger    	shift = length - aRange.length;
+  NSUInteger	need = size + shift;
   void		*buf;
 
   if (aRange.location > size)
@@ -2085,6 +2110,11 @@ failure:
       [self setLength: need];
     }
   buf = [self mutableBytes];
+  if (0 == buf)
+    {
+      [NSException raise: NSInternalInconsistencyException
+		  format: @"missing bytes in replaceByteInRange:withBytes:"];
+    }
   if (shift < 0)
     {
       if (length > 0)
@@ -2119,10 +2149,16 @@ failure:
  */
 - (void) resetBytesInRange: (NSRange)aRange
 {
-  unsigned	size = [self length];
+  NSUInteger	size = [self length];
+  void		*bytes = [self mutableBytes];
 
   GS_RANGE_CHECK(aRange, size);
-  memset((char*)[self bytes] + aRange.location, 0, aRange.length);
+  if (0 == bytes)
+    {
+      [NSException raise: NSInternalInconsistencyException
+		  format: @"missing bytes in resetBytesInRange:"];
+    }
+  memset((char*)bytes + aRange.location, 0, aRange.length);
 }
 
 /**
@@ -2840,7 +2876,7 @@ getBytes(void* dst, void* src, unsigned len, unsigned limit, unsigned *pos)
   if (*cursor >= length)
     {
       [NSException raise: NSRangeException
-		  format: @"Range: (%u, 1) Size: %d", *cursor, length];
+		  format: @"Range: (%u, 1) Size: %"PRIuPTR, *cursor, length];
     }
   *tag = *((unsigned char*)bytes + (*cursor)++);
   if (*tag & _GSC_MAYX)
@@ -2856,7 +2892,7 @@ getBytes(void* dst, void* src, unsigned len, unsigned limit, unsigned *pos)
 	      if (*cursor >= length)
 		{
 		  [NSException raise: NSRangeException
-			      format: @"Range: (%u, 1) Size: %d",
+			      format: @"Range: (%u, 1) Size: %"PRIuPTR,
 			  *cursor, length];
 		}
 	      *ref = (unsigned int)*((unsigned char*)bytes + (*cursor)++);
@@ -2869,7 +2905,7 @@ getBytes(void* dst, void* src, unsigned len, unsigned limit, unsigned *pos)
 	      if (*cursor >= length-1)
 		{
 		  [NSException raise: NSRangeException
-			      format: @"Range: (%u, 1) Size: %d",
+			      format: @"Range: (%u, 1) Size: %"PRIuPTR,
 			  *cursor, length];
 		}
 #if NEED_WORD_ALIGNMENT
@@ -2889,7 +2925,7 @@ getBytes(void* dst, void* src, unsigned len, unsigned limit, unsigned *pos)
 	      if (*cursor >= length-3)
 		{
 		  [NSException raise: NSRangeException
-			      format: @"Range: (%u, 1) Size: %d",
+			      format: @"Range: (%u, 1) Size: %"PRIuPTR,
 			  *cursor, length];
 		}
 #if NEED_WORD_ALIGNMENT
@@ -3054,7 +3090,7 @@ getBytes(void* dst, void* src, unsigned len, unsigned limit, unsigned *pos)
   bytes = mmap(0, length, PROT_READ, MAP_SHARED, fd, 0);
   if (bytes == MAP_FAILED)
     {
-      NSWarnMLog(@"mapping failed for %s - %@", path, [NSError _last]);
+      NSWarnMLog(@"mapping failed for %@ - %@", path, [NSError _last]);
       close(fd);
       DESTROY(self);
       self = [dataMalloc allocWithZone: NSDefaultMallocZone()];
@@ -3342,8 +3378,8 @@ getBytes(void* dst, void* src, unsigned len, unsigned limit, unsigned *pos)
 {
   if (bufferSize > 0)
     {
-      unsigned	oldLength = length;
-      unsigned	minimum = length + bufferSize;
+      NSUInteger oldLength = length;
+      NSUInteger minimum = length + bufferSize;
 
       if (aBuffer == 0)
 	{
@@ -3369,12 +3405,12 @@ getBytes(void* dst, void* src, unsigned len, unsigned limit, unsigned *pos)
 {
   if (minimum > capacity)
     {
-      unsigned	nextCapacity = capacity + growth;
-      unsigned	nextGrowth = capacity ? capacity : 1;
+      NSUInteger nextCapacity = capacity + growth;
+      NSUInteger nextGrowth = capacity ? capacity : 1;
 
       while (nextCapacity < minimum)
 	{
-	  unsigned	tmp = nextCapacity + nextGrowth;
+	  NSUInteger tmp = nextCapacity + nextGrowth;
 
 	  nextGrowth = nextCapacity;
 	  nextCapacity = tmp;
@@ -3392,7 +3428,7 @@ getBytes(void* dst, void* src, unsigned len, unsigned limit, unsigned *pos)
 - (void) replaceBytesInRange: (NSRange)aRange
 		   withBytes: (const void*)moreBytes
 {
-  unsigned	need = NSMaxRange(aRange);
+  NSUInteger need = NSMaxRange(aRange);
 
   if (aRange.location > length)
     {
@@ -3724,7 +3760,7 @@ getBytes(void* dst, void* src, unsigned len, unsigned limit, unsigned *pos)
       if (tmp == 0)
 	{
 	  [NSException raise: NSMallocException
-	    format: @"Unable to set data capacity to '%d'", size];
+	    format: @"Unable to set data capacity to '%"PRIuPTR"'", size];
 	}
       if (bytes)
 	{
@@ -3760,7 +3796,7 @@ getBytes(void* dst, void* src, unsigned len, unsigned limit, unsigned *pos)
 
 - (void) setData: (NSData*)data
 {
-  unsigned	l = [data length];
+  NSUInteger l = [data length];
 
   [self setCapacity: l];
   length = l;
@@ -3771,7 +3807,13 @@ getBytes(void* dst, void* src, unsigned len, unsigned limit, unsigned *pos)
 {
   if (size > capacity)
     {
-      [self setCapacity: size];
+      NSUInteger    growTo = capacity + capacity / 2;
+
+      if (size > growTo)
+        {
+          growTo = size;
+        }
+      [self setCapacity: growTo];
     }
   if (size > length)
     {
@@ -3912,8 +3954,8 @@ getBytes(void* dst, void* src, unsigned len, unsigned limit, unsigned *pos)
       if (newid == -1)			/* Created memory? */
 	{
 	  [NSException raise: NSMallocException
-	    format: @"Unable to create shared memory segment (size:%u) - %@.",
-	    size, [NSError _last]];
+	    format: @"Unable to create shared memory segment"
+	    @" (size:%"PRIuPTR") - %@.", size, [NSError _last]];
 	}
       tmp = shmat(newid, 0, 0);
       if ((intptr_t)tmp == -1)			/* Attached memory? */
