@@ -24,7 +24,7 @@
    Boston, MA 02111 USA.
 
    <title>NSDebug utilities reference</title>
-   $Date: 2011-12-15 02:42:39 -0700 (Thu, 15 Dec 2011) $ $Revision: 34290 $
+   $Date: 2013-08-22 16:44:54 +0100 (Thu, 22 Aug 2013) $ $Revision: 37003 $
    */
 
 #import "common.h"
@@ -87,6 +87,7 @@ static void (*_GSDebugAllocationRemoveFunc)(Class c, id o)
 + (void) initialize
 {
   uniqueLock = [GSLazyRecursiveLock new];
+  [[NSObject leakAt: &uniqueLock] release];
 }
 @end
 
@@ -525,27 +526,30 @@ _GSDebugAllocationList(BOOL difference)
       buf = NSZoneMalloc(NSDefaultMallocZone(), siz);
     }
 
-  if (buf)
+  if (0 == buf)
     {
-      pos = 0;
-      for (i = 0; i < num_classes; i++)
+      return "";
+    }
+
+  pos = 0;
+  for (i = 0; i < num_classes; i++)
+    {
+      int	val = the_table[i].count;
+
+      if (difference)
 	{
-	  int	val = the_table[i].count;
+	  val -= the_table[i].lastc;
+	}
+      the_table[i].lastc = the_table[i].count;
 
-	  if (difference)
-	    {
-	      val -= the_table[i].lastc;
-	    }
-	  the_table[i].lastc = the_table[i].count;
-
-	  if (val != 0)
-	    {
-	      snprintf(&buf[pos], siz - pos, "%d\t%s\n",
-		val, class_getName(the_table[i].class));
-	      pos += strlen(&buf[pos]);
-	    }
+      if (val != 0)
+	{
+	  snprintf(&buf[pos], siz - pos, "%d\t%s\n",
+	    val, class_getName(the_table[i].class));
+	  pos += strlen(&buf[pos]);
 	}
     }
+
   return buf;
 }
 
@@ -612,21 +616,24 @@ _GSDebugAllocationListAll(void)
       buf = NSZoneMalloc(NSDefaultMallocZone(), siz);
     }
 
-  if (buf)
+  if (0 == buf)
     {
-      pos = 0;
-      for (i = 0; i < num_classes; i++)
-	{
-	  int	val = the_table[i].total;
+      return "";
+    }
 
-	  if (val != 0)
-	    {
-	      snprintf(&buf[pos], siz - pos, "%d\t%s\n",
-		val, class_getName(the_table[i].class));
-	      pos += strlen(&buf[pos]);
-	    }
+  pos = 0;
+  for (i = 0; i < num_classes; i++)
+    {
+      int	val = the_table[i].total;
+
+      if (val != 0)
+	{
+	  snprintf(&buf[pos], siz - pos, "%d\t%s\n",
+	    val, class_getName(the_table[i].class));
+	  pos += strlen(&buf[pos]);
 	}
     }
+
   return buf;
 }
 
