@@ -24,7 +24,7 @@
    Boston, MA 02111 USA.
 
    <title>NSHost class reference</title>
-   $Date: 2012-01-25 12:52:34 -0700 (Wed, 25 Jan 2012) $ $Revision: 34629 $
+   $Date: 2013-08-22 16:44:54 +0100 (Thu, 22 Aug 2013) $ $Revision: 37003 $
   */
 
 #import "common.h"
@@ -191,6 +191,7 @@ static id			null = nil;
 	    {
 	      NSString	*addr;
 
+	      memset((void*)&in, '\0', sizeof(in));
 	      memcpy((void*)&in.s_addr, (const void*)ptr, entry->h_length);
 	      addr = [NSString stringWithUTF8String: (char*)inet_ntoa(in)];
 	      [addresses addObject: addr];
@@ -264,8 +265,11 @@ myHostName()
     {
       hostClass = self;
       null = [[NSNull null] retain];
+      [[NSObject leakAt: &null] release];
       _hostCacheLock = [[NSRecursiveLock alloc] init];
+      [[NSObject leakAt: &_hostCacheLock] release];
       _hostCache = [NSMutableDictionary new];
+      [[NSObject leakAt: &_hostCache] release];
     }
 }
 
