@@ -23,7 +23,7 @@
    Boston, MA 02111 USA.
 
    <title>NSInvocation class reference</title>
-   $Date: 2012-01-29 09:57:06 -0700 (Sun, 29 Jan 2012) $ $Revision: 34660 $
+   $Date: 2013-08-22 16:44:54 +0100 (Thu, 22 Aug 2013) $ $Revision: 37003 $
    */
 
 #import "common.h"
@@ -35,7 +35,6 @@
 #import "Foundation/NSZone.h"
 #import "GSInvocation.h"
 #import "GSPrivate.h"
-#import "GNUstepBase/NSObject+GNUstepBase.h"
 
 #if defined(USE_LIBFFI)
 #include "cifframe.h"
@@ -71,6 +70,7 @@
 
 - (void) dealloc
 {
+  DESTROY(frame);
   if (size > 0)
     {
 #if	defined(HAVE_FFI_PREP_CLOSURE_LOC)
@@ -173,6 +173,11 @@
 #endif
 #endif
 }
+
+- (void) setFrame: (id)aFrame
+{
+  ASSIGN(frame, aFrame);
+}
 @end
 
 static Class   NSInvocation_abstract_class;
