@@ -7,10 +7,25 @@
  * data for the parse can be either a static JSON string or some JSON data.
  */
 
-#import <Foundation/Foundation.h>
-#import <GNUstepBase/NSObject+GNUstepBase.h>
+#import "common.h"
+#import "Foundation/NSArray.h"
+#import "Foundation/NSByteOrder.h"
+#import "Foundation/NSCharacterSet.h"
+#import "Foundation/NSData.h"
+#import "Foundation/NSDictionary.h"
+#import "Foundation/NSError.h"
+#import "Foundation/NSJSONSerialization.h"
+#import "Foundation/NSNull.h"
+#import "Foundation/NSStream.h"
+#import "Foundation/NSString.h"
+#import "Foundation/NSValue.h"
 #import "GSFastEnumeration.h"
 
+/* Boolean constants.
+ */
+static id       boolN;
+static id       boolY;
+
 /**
  * The number of (unicode) characters to fetch from the source at once.
  */
@@ -260,7 +275,7 @@ parseError(ParserState *state)
    */
   NSDictionary *userInfo = [[NSDictionary alloc] initWithObjectsAndKeys:
     _(@"JSON Parse error"), NSLocalizedDescriptionKey,
-    _(([NSString stringWithFormat: @"Unexpected character %c at index %d",
+    _(([NSString stringWithFormat: @"Unexpected character %c at index %"PRIdPTR,
         (char)currentChar(state), state->sourceIndex])), 
       NSLocalizedFailureReasonErrorKey,
     nil];
@@ -556,7 +571,7 @@ parseObject(ParserState *state)
       c = consumeSpace(state);
       if (c == ',')
         {
-          c = consumeChar(state);
+          consumeChar(state);
         }
       c = consumeSpace(state);
     }
@@ -597,10 +612,11 @@ parseValue(ParserState *state)
       // Literal null
       case 'n':
         {
-          if ((consumeChar(state) == 'u') &&
-              (consumeChar(state) == 'l') &&
-              (consumeChar(state) == 'l'))
+          if ((consumeChar(state) == 'u')
+	    && (consumeChar(state) == 'l')
+	    && (consumeChar(state) == 'l'))
             {
+	      consumeChar(state);
               return [[NSNull null] retain];
             }
           break;
@@ -608,22 +624,24 @@ parseValue(ParserState *state)
       // literal 
       case 't':
         {
-          if ((consumeChar(state) == 'r') &&
-              (consumeChar(state) == 'u') &&
-              (consumeChar(state) == 'e'))
+          if ((consumeChar(state) == 'r')
+	    && (consumeChar(state) == 'u')
+	    && (consumeChar(state) == 'e'))
             {
-              return [[NSNumber alloc] initWithBool: YES];
+	      consumeChar(state);
+              return [boolY retain];
             }
           break;
         }
       case 'f':
         {
-          if ((consumeChar(state) == 'a') &&
-              (consumeChar(state) == 'l') &&
-              (consumeChar(state) == 's') &&
-              (consumeChar(state) == 'e'))
+          if ((consumeChar(state) == 'a')
+	    && (consumeChar(state) == 'l')
+	    && (consumeChar(state) == 's')
+	    && (consumeChar(state) == 'e'))
             {
-              return [[NSNumber alloc] initWithBool: NO];
+	      consumeChar(state);
+              return [boolN retain];
             }
           break;
         }
@@ -707,10 +725,13 @@ getEncoding(const uint8_t BOM[4], ParserState *state)
 /**
  * Classes that are permitted to be written.  
  */
-static Class NSNullClass, NSArrayClass, NSStringClass, NSDictionaryClass,
-             NSNumberClass;
+static Class NSArrayClass;
+static Class NSDictionaryClass;
+static Class NSNullClass;
+static Class NSNumberClass;
+static Class NSStringClass;
 
-static NSCharacterSet *escapeSet;
+static NSMutableCharacterSet *escapeSet;
 
 static inline void
 writeTabs(NSMutableString *output, NSInteger tabs)
@@ -777,56 +798,110 @@ writeObject(id obj, NSMutableString *output, NSInteger tabs)
     }
   else if ([obj isKindOfClass: NSStringClass])
     {
-      NSRange r = [obj rangeOfCharacterFromSet: escapeSet];
-      if (r.location != NSNotFound)
-        {
-          NSMutableString *str = [obj mutableCopy];
-          NSCharacterSet *controlSet = [NSCharacterSet controlCharacterSet];
-          [str replaceOccurrencesOfString: @"\\"
-                               withString: @"\\\\"
-                                  options: 0
-                                    range: NSMakeRange(0, [str length])];
-          [str replaceOccurrencesOfString: @"\""
-                               withString: @"\\\""
-                                  options: 0
-                                    range: NSMakeRange(0, [str length])];
-          r = [str rangeOfCharacterFromSet: controlSet];
-          while (r.location != NSNotFound)
-            {
-              unichar control = [str characterAtIndex: r.location];
-              NSString *escaped;
+      NSString  *str = (NSString*)obj;
+      unsigned	length = [str length];
 
-              escaped = [[NSString alloc] initWithFormat: @"\\u%.4d",
-		(int)control];
-              [str replaceCharactersInRange: r
-                                 withString: escaped];
-              [escaped release];
-              r = [str rangeOfCharacterFromSet: controlSet];
-            }
-          [output appendFormat: @"\"%@\"", str];
-          [str release];
+      if (length == 0)
+        {
+          [output appendString: @"\"\""];
         }
       else
         {
-          [output appendFormat: @"\"%@\"", obj];
+          unsigned	size = 2;
+          unichar	*from;
+          unsigned	i = 0;
+          unichar	*to;
+          unsigned	j = 0;
+
+          from = NSZoneMalloc (NSDefaultMallocZone(), sizeof(unichar) * length);
+          [str getCharacters: from];
+
+          for (i = 0; i < length; i++)
+            {
+              unichar	c = from[i];
+
+              if (c == '"' || c == '\\' || c == '\b'
+                || c == '\f' || c == '\n' || c == '\r' || c == '\t')
+                {
+                  size += 2;
+                }
+              else if (c < 0x20)
+                {
+                  size += 6;
+                }
+              else
+                {
+                  size++;
+                }
+            }
+
+          to = NSZoneMalloc (NSDefaultMallocZone(), sizeof(unichar) * size);
+          to[j++] = '"';
+          for (i = 0; i < length; i++)
+            {
+              unichar	c = from[i];
+
+              if (c == '"' || c == '\\' || c == '\b'
+                || c == '\f' || c == '\n' || c == '\r' || c == '\t')
+                {
+                  to[j++] = '\\';
+                  switch (c)
+                    {
+                      case '\\': to[j++] = '\\'; break;
+                      case '\b': to[j++] = 'b'; break;
+                      case '\f': to[j++] = 'f'; break;
+                      case '\n': to[j++] = 'n'; break;
+                      case '\r': to[j++] = 'r'; break;
+                      case '\t': to[j++] = 't'; break;
+                      default: to[j++] = '"'; break;
+                    }
+                }
+              else if (c < 0x20)
+                {
+                  char	buf[5];
+
+                  to[j++] = '\\';
+                  to[j++] = 'u';
+                  sprintf(buf, "%04x", c);
+                  to[j++] = buf[0];
+                  to[j++] = buf[1];
+                  to[j++] = buf[2];
+                  to[j++] = buf[3];
+                }
+              else
+                {
+                  to[j++] = c;
+                }
+            }
+          to[j] = '"';
+          str = [[NSStringClass alloc] initWithCharacters: to length: size];
+          NSZoneFree (NSDefaultMallocZone (), to);
+          NSZoneFree (NSDefaultMallocZone (), from);
+          [output appendString: str];
+          [str release];
         }
     }
+  else if (obj == boolN)
+    {
+      [output appendString: @"false"];
+    }
+  else if (obj == boolY)
+    {
+      [output appendString: @"true"];
+    }
   else if ([obj isKindOfClass: NSNumberClass])
     {
-      if ([obj objCType][0] == @encode(BOOL)[0])
+      const char        *t = [obj objCType];
+
+      if (strchr("cCsSiIlLqQ", *t) != 0)
         {
-          if ([obj boolValue])
-            {
-              [output appendString: @"true"];
-            }
-          else
-            {
-              [output appendString: @"false"];
-            }
+          long long     i = [(NSNumber*)obj longLongValue];
+
+          [output appendFormat: @"%lld", i];
         }
       else
         {
-          [output appendFormat: @"%f", [obj doubleValue]];
+          [output appendFormat: @"%g", [(NSNumber*)obj doubleValue]];
         }
     }
   else if ([obj isKindOfClass: NSNullClass])
@@ -843,14 +918,24 @@ writeObject(id obj, NSMutableString *output, NSInteger tabs)
 @implementation NSJSONSerialization
 + (void) initialize
 {
-  NSNullClass = [NSNull class];
-  NSArrayClass = [NSArray class];
-  NSStringClass = [NSString class];
-  NSDictionaryClass = [NSDictionary class];
-  NSNumberClass = [NSNumber class];
-  escapeSet
-    = [[NSCharacterSet characterSetWithCharactersInString: @"\"\\"] retain];
+  static BOOL beenHere = NO;
 
+  if (NO == beenHere)
+    {
+      NSNullClass = [NSNull class];
+      NSArrayClass = [NSArray class];
+      NSStringClass = [NSString class];
+      NSDictionaryClass = [NSDictionary class];
+      NSNumberClass = [NSNumber class];
+      escapeSet = [NSMutableCharacterSet new];
+      [[NSObject leakAt: &escapeSet] release];
+      [escapeSet addCharactersInString: @"\"\\"];
+      boolN = [[NSNumber alloc] initWithBool: NO];
+      [[NSObject leakAt: &boolN] release];
+      boolY = [[NSNumber alloc] initWithBool: YES];
+      [[NSObject leakAt: &boolY] release];
+      beenHere = YES;
+    }
 }
 
 + (NSData*) dataWithJSONObject: (id)obj
