@@ -21,7 +21,7 @@
    Boston, MA 02111 USA.
 
    <title>NSKeyValueCoding informal protocol reference</title>
-   $Date: 2012-01-30 04:31:40 -0700 (Mon, 30 Jan 2012) $ $Revision: 34665 $
+   $Date: 2013-07-02 16:46:26 +0100 (Tue, 02 Jul 2013) $ $Revision: 36799 $
    */
 
 #import "common.h"
@@ -342,9 +342,9 @@ static id ValueForKey(NSObject *self, const char *key, unsigned size)
     }
 #endif
   [NSException raise: NSInvalidArgumentException
-    format: @"%@ -- %@ 0x%x: Given nil value to set for key \"%@\"",
+    format: @"%@ -- %@ 0x%"PRIxPTR": Given nil value to set for key \"%@\"",
     NSStringFromSelector(_cmd), NSStringFromClass([self class]),
-    self, aKey];
+    (NSUInteger)self, aKey];
 }
 
 
@@ -898,9 +898,10 @@ static id ValueForKey(NSObject *self, const char *key, unsigned size)
 - (void) unableToSetNilForKey: (NSString*)aKey
 {
   GSOnceMLog(@"This method is deprecated, use -setNilValueForKey:");
-  [NSException raise: NSInvalidArgumentException
-	      format: @"%@ -- %@ 0x%x: Given nil value to set for key \"%@\"",
-    NSStringFromSelector(_cmd), NSStringFromClass([self class]), self, aKey];
+  [NSException raise: NSInvalidArgumentException format:
+    @"%@ -- %@ 0x%"PRIxPTR": Given nil value to set for key \"%@\"",
+    NSStringFromSelector(_cmd), NSStringFromClass([self class]),
+    (NSUInteger)self, aKey];
 }
 
 
