@@ -22,7 +22,7 @@
    Boston, MA 02111 USA.
 
    <title>NSLog reference</title>
-   $Date: 2011-10-14 09:46:13 -0600 (Fri, 14 Oct 2011) $ $Revision: 33993 $
+   $Date: 2013-07-03 07:46:41 +0100 (Wed, 03 Jul 2013) $ $Revision: 36810 $
    */
 
 #import "common.h"
@@ -39,6 +39,16 @@
 
 #ifdef	HAVE_SYSLOG_H
 #include <syslog.h>
+#elif HAVE_SYS_SLOG_H
+  // we are on a QNX-ish system, which has a syslog symbol somewhere, but it isn't
+  // our syslog function (we use slogf().)
+# ifdef HAVE_SYSLOG
+#   undef HAVE_SYSLOG
+# endif
+# include <sys/slog.h>
+# ifdef HAVE_SYS_SLOGCODES_H
+#   include <sys/slogcodes.h>
+# endif
 #endif
 
 #define	UNISTR(X) \
@@ -98,7 +108,7 @@ GSLogLock()
 }
 
 static void
-_NSLog_standard_printf_handler (NSString* message)
+_NSLog_standard_printf_handler(NSString* message)
 {
   NSData	*d;
   const char	*buf;
@@ -106,8 +116,8 @@ _NSLog_standard_printf_handler (NSString* message)
 #if	defined(__MINGW__)
   LPCWSTR	null_terminated_buf;
 #else
-#if	defined(HAVE_SYSLOG)
-  char	*null_terminated_buf;
+#if	defined(HAVE_SYSLOG) || defined(HAVE_SLOGF)
+  char	*null_terminated_buf = NULL;
 #endif
 #endif
   static NSStringEncoding enc = 0;
@@ -162,8 +172,8 @@ _NSLog_standard_printf_handler (NSString* message)
 	    NULL);			// pointer to data
 	}
     }
-#else      
-      
+#else
+
 #if	defined(HAVE_SYSLOG)
   if (GSPrivateDefaultsFlag(GSLogSyslog) == YES
     || write(_NSLogDescriptor, buf, len) != (int)len)
@@ -176,6 +186,40 @@ _NSLog_standard_printf_handler (NSString* message)
 
       free(null_terminated_buf);
     }
+#elif defined(HAVE_SLOGF)
+  if (GSPrivateDefaultsFlag(GSLogSyslog) == YES
+    || write(_NSLogDescriptor, buf, len) != (int)len)
+    {
+      /* QNX's slog has a size limit per entry. We might need to iterate over
+       * _SLOG_MAXSIZEd chunks of the buffer
+       */
+      const char *newBuf = buf;
+      unsigned newLen = len;
+
+      // Allocate at most _SLOG_MAXSIZE bytes
+      null_terminated_buf = malloc(sizeof(char) * MIN(newLen, _SLOG_MAXSIZE));
+      // If it's shorter than that, we never even enter the loop
+      while (newLen >= _SLOG_MAXSIZE)
+        {
+          strncpy(null_terminated_buf, newBuf, (_SLOG_MAXSIZE - 1));
+          null_terminated_buf[_SLOG_MAXSIZE] = '\0';
+          slogf(_SLOG_SETCODE(_SLOG_SYSLOG, 0), _SLOG_ERROR, "%s",
+            null_terminated_buf);
+          newBuf += (_SLOG_MAXSIZE - 1);
+          newLen -= (_SLOG_MAXSIZE - 1);
+        }
+      /* Write out the rest (which will be at most (_SLOG_MAXSIZE - 1) chars,
+       * so the terminator still fits.
+       */
+      if (0 != newLen)
+        {
+          strncpy(null_terminated_buf, newBuf, newLen);
+          null_terminated_buf[newLen] = '\0';
+          slogf(_SLOG_SETCODE(_SLOG_SYSLOG, 0), _SLOG_ERROR, "%s",
+            null_terminated_buf);
+        }
+      free(null_terminated_buf);
+    }
 #else
   write(_NSLogDescriptor, buf, len);
 #endif
@@ -239,13 +283,13 @@ NSLog_printf_handler *_NSLog_printf_handler = _NSLog_standard_printf_handler;
  * </p>
  */
 void
-NSLog (NSString* format, ...)
+NSLog(NSString* format, ...)
 {
   va_list ap;
 
-  va_start (ap, format);
-  NSLogv (format, ap);
-  va_end (ap);
+  va_start(ap, format);
+  NSLogv(format, ap);
+  va_end(ap);
 }
 
 /**
@@ -275,7 +319,7 @@ NSLog (NSString* format, ...)
  * </p>
  */
 void
-NSLogv (NSString* format, va_list args)
+NSLogv(NSString* format, va_list args)
 {
   NSString		*prefix;
   NSString		*message;
@@ -301,8 +345,8 @@ NSLogv (NSString* format, va_list args)
     {
       if (GSPrivateDefaultsFlag(GSLogThread) == YES)
 	{
-	  prefix = [NSString stringWithFormat: @"[thread:%x] ",
-	    GSCurrentThread()];
+	  prefix = [NSString stringWithFormat: @"[thread:%"PRIxPTR"] ",
+	    (NSUInteger)GSCurrentThread()];
 	}
       else
 	{
@@ -315,11 +359,11 @@ NSLogv (NSString* format, va_list args)
       if (GSPrivateDefaultsFlag(GSLogThread) == YES)
 	{
 	  prefix = [NSString
-	    stringWithFormat: @"%@ %@[%d,%x] ",
+	    stringWithFormat: @"%@ %@[%d,%"PRIxPTR"x] ",
 	    [[NSCalendarDate calendarDate]
 	      descriptionWithCalendarFormat: @"%Y-%m-%d %H:%M:%S.%F"],
 	    [[NSProcessInfo processInfo] processName],
-	    pid, GSCurrentThread()];
+	    pid, (NSUInteger)GSCurrentThread()];
 	}
       else
 	{
