@@ -4,19 +4,19 @@
    Written by:  Gregory Casamento <greg.casamento@gmail.com>
    Written by:  Richard Frith-Macdonald <rfm@gnu.org>
    Date: 2009,2010
-   
+
    This file is part of the GNUstep Base Library.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
-   
+
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
-   
+
    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
@@ -24,7 +24,7 @@
 
    <title>NSOperation class reference</title>
    $Date: 2008-06-08 11:38:33 +0100 (Sun, 08 Jun 2008) $ $Revision: 26606 $
-   */ 
+   */
 
 #import "common.h"
 
@@ -41,7 +41,8 @@
   BOOL finished; \
   BOOL blocked; \
   BOOL ready; \
-  NSMutableArray *dependencies;
+  NSMutableArray *dependencies; \
+  GSOperationCompletionBlock completionBlock;
 
 #define	GS_NSOperationQueue_IVARS \
   NSRecursiveLock	*lock; \
@@ -91,6 +92,7 @@ static NSArray	*empty = nil;
 + (void) initialize
 {
   empty = [NSArray new];
+  [[NSObject leakAt: &empty] release];
 }
 
 - (void) addDependency: (NSOperation *)op
@@ -188,6 +190,11 @@ static NSArray	*empty = nil;
     }
 }
 
+- (GSOperationCompletionBlock) completionBlock
+{
+  return internal->completionBlock;
+}
+
 - (void) dealloc
 {
   if (internal != nil)
@@ -353,6 +360,11 @@ static NSArray	*empty = nil;
   [internal->lock unlock];
 }
 
+- (void) setCompletionBlock: (GSOperationCompletionBlock)aBlock
+{
+  internal->completionBlock = aBlock;
+}
+
 - (void) setQueuePriority: (NSOperationQueuePriority)pri
 {
   if (pri <= NSOperationQueuePriorityVeryLow)
@@ -524,6 +536,10 @@ static NSArray	*empty = nil;
 	  internal->finished = YES;
 	  [self didChangeValueForKey: @"isFinished"];
 	}
+      if (NULL != internal->completionBlock)
+	{
+	  CALL_BLOCK_NO_ARGS(internal->completionBlock);
+	}
     }
   [internal->lock unlock];
   [self release];
@@ -554,7 +570,7 @@ sortFunc(id o1, id o2, void *ctxt)
 {
   NSOperationQueuePriority p1 = [o1 queuePriority];
   NSOperationQueuePriority p2 = [o2 queuePriority];
-  
+
   if (p1 < p2) return NSOrderedDescending;
   if (p1 > p2) return NSOrderedAscending;
   return NSOrderedSame;
@@ -703,7 +719,7 @@ static NSOperationQueue *mainQueue = nil;
       if (YES == invalidArg)
 	{
 	  [NSException raise: NSInvalidArgumentException
-	    format: @"[%@-%@] object at index %u is not an NSOperation",
+	    format: @"[%@-%@] object at index %"PRIuPTR" is not an NSOperation",
 	    NSStringFromClass([self class]), NSStringFromSelector(_cmd),
 	    index];
 	}
@@ -798,7 +814,7 @@ static NSOperationQueue *mainQueue = nil;
     && cnt != NSOperationQueueDefaultMaxConcurrentOperationCount)
     {
       [NSException raise: NSInvalidArgumentException
-		  format: @"[%@-%@] cannot set negative (%d) count",
+		  format: @"[%@-%@] cannot set negative (%"PRIdPTR") count",
 	NSStringFromClass([self class]), NSStringFromSelector(_cmd), cnt];
     }
   [internal->lock lock];
@@ -863,6 +879,7 @@ static NSOperationQueue *mainQueue = nil;
   if (mainQueue == nil)
     {
       mainQueue = [self new];
+      [[NSObject leakAt: &mainQueue] release];
       [[[NSThread currentThread] threadDictionary] setObject: mainQueue
 						      forKey: threadKey];
     }
