@@ -26,7 +26,7 @@
    Boston, MA 02111 USA.
 
    <title>NSPathUtilities function reference</title>
-   $Date: 2011-11-14 03:01:24 -0700 (Mon, 14 Nov 2011) $ $Revision: 34169 $
+   $Date: 2013-08-22 16:44:54 +0100 (Thu, 22 Aug 2013) $ $Revision: 37003 $
    */
 
 /**
@@ -223,6 +223,9 @@ static BOOL ParseConfigurationFile(NSString *name, NSMutableDictionary *dict,
 static void InitialisePathUtilities(void);
 static void ShutdownPathUtilities(void);
 
+@interface GSPathUtilities : NSObject
+@end
+
 /* Conditionally assign an object from a dictionary to var
  * We don't need to retain val before releasing var, because we
  * can be sure that if var is val it is retained by the dictionary
@@ -269,9 +272,11 @@ static void ShutdownPathUtilities(void);
     }\
 })
 
-/* Conditionally assign lval to var only if var is nil */
+/* Conditionally assign lval to var only if var is nil
+
+lval must be casted because Clang disallows ObjC literals comparison */
 #define TEST_ASSIGN(var, lval) ({\
-  if ((var == nil)&&(lval != nil))\
+  if ((var == nil)&&((NSString *)lval != nil))\
     {\
       var = RETAIN(lval);\
     }\
@@ -680,7 +685,10 @@ static void ExtractValuesFromConfig(NSDictionary *config)
       if ([manager fileExistsAtPath: path isDirectory: &flag] == NO
 	|| flag == NO)
 	{
-	  [manager createDirectoryAtPath: path attributes: attr];
+	  [manager createDirectoryAtPath: path
+             withIntermediateDirectories: YES
+                              attributes: attr
+                                   error: NULL];
 	}
     }
 
@@ -889,10 +897,16 @@ NSMutableDictionary*
 GNUstepConfig(NSDictionary *newConfig)
 {
   static NSDictionary	*config = nil;
+  static BOOL beenHere = NO;
   NSMutableDictionary	*conf = nil;
   BOOL			changedSystemConfig = NO;
 
   [gnustep_global_lock lock];
+  if (NO == beenHere)
+    {
+      beenHere = YES;
+      [[NSObject leakAt: &config] release];
+    }
   if (config == nil || (newConfig != nil && [config isEqual: newConfig] == NO))
     {
       NS_DURING
@@ -1119,8 +1133,18 @@ static void InitialisePathUtilities(void)
   NS_DURING
     {
       NSMutableDictionary	*config;
+      static BOOL               beenHere = NO;
 
       [gnustep_global_lock lock];
+      if (NO == beenHere)
+        {
+          beenHere = YES;
+          if (YES == [NSObject shouldCleanUp])
+            {
+              // Get path utilities shutdown called at process exit.
+              [GSPathUtilities class];
+            }
+        }
       ASSIGNCOPY(uninstalled, [[[NSProcessInfo processInfo] environment]
 	objectForKey: @"GNUSTEP_UNINSTALLED_LIBRARY_DIRECTORY"]);
       gnustepUserName = [NSUserName() copy];
@@ -1803,6 +1827,7 @@ NSFullUserName(void)
 #else
 #ifdef  HAVE_PWD_H
 #if     defined(HAVE_GETPWNAM_R)
+#if     defined(HAVE_PW_GECOS_IN_PASSWD)
       struct passwd pw;
       struct passwd *p;
       char buf[BUFSIZ*10];
@@ -1814,8 +1839,10 @@ NSFullUserName(void)
               userName = [NSString stringWithUTF8String: pw.pw_gecos];
 	    }
         }
+#endif /* HAVE_PW_GECOS_IN_PASSWD */
 #else
 #if     defined(HAVE_GETPWNAM)
+#if     defined(HAVE_PW_GECOS_IN_PASSWD)
       struct passwd	*pw;
 
       [gnustep_global_lock lock];
@@ -1825,6 +1852,7 @@ NSFullUserName(void)
           userName = [NSString stringWithUTF8String: pw->pw_gecos];
         }
       [gnustep_global_lock lock];
+#endif /* HAVE_PW_GECOS_IN_PASSWD */
 #endif /* HAVE_GETPWNAM */
 #endif /* HAVE_GETPWNAM_R */
 #endif /* HAVE_PWD_H */
@@ -1842,7 +1870,6 @@ NSFullUserName(void)
 NSString *
 GSDefaultsRootForUser(NSString *userName)
 {
-  NSString *home;
   NSString *defaultsDir;
 
   InitialisePathUtilities();
@@ -1850,7 +1877,6 @@ GSDefaultsRootForUser(NSString *userName)
     {
       userName = NSUserName();
     }
-  home = NSHomeDirectoryForUser(userName);
   if ([userName isEqual: NSUserName()])
     {
       defaultsDir = gnustepUserDefaultsDir;
@@ -1873,9 +1899,14 @@ GSDefaultsRootForUser(NSString *userName)
       return defaultsDir;	// Just use windows registry.
     }
 #endif
-  home = [home stringByAppendingPathComponent: defaultsDir];
+  if (NO == [defaultsDir isAbsolutePath])
+    {
+      NSString  *home = NSHomeDirectoryForUser(userName);
 
-  return home;
+      defaultsDir = [home stringByAppendingPathComponent: defaultsDir];
+    }
+
+  return defaultsDir;
 }
 
 NSArray *
@@ -1930,6 +1961,7 @@ NSTemporaryDirectory(void)
 	  if (baseTempDirName == nil)
 	    {
 #if	defined(__CYGWIN__)
+#warning Basing temporary directory in /cygdrive/c; any reason?
 	      baseTempDirName = @"/cygdrive/c/";
 #elif	defined(__MINGW__)
 	      baseTempDirName = @"C:\\";
@@ -1970,7 +2002,9 @@ NSTemporaryDirectory(void)
   perm = perm & 0777;
 
 // Mateu Batle: secure temporary directories don't work in MinGW
-#ifndef __MINGW__
+// Ivan Vucica: there are also problems with Cygwin
+//              probable cause: http://stackoverflow.com/q/9561759/39974
+#if !defined(__MINGW__) && !defined(__CYGWIN__)
 
 #if	defined(__MINGW__)
   uid = owner;
@@ -1984,6 +2018,10 @@ NSTemporaryDirectory(void)
   if ((perm != 0700 && perm != 0600) || owner != uid)
     {
       NSString	*secure;
+      NSNumber	*p = [NSNumber numberWithInt: 0700];
+
+      attr = [NSDictionary dictionaryWithObject: p
+                                         forKey: NSFilePosixPermissions];
 
       /*
        * The name of the secure subdirectory reflects the user ID rather
@@ -1998,18 +2036,20 @@ NSTemporaryDirectory(void)
 
       if ([manager fileExistsAtPath: tempDirName] == NO)
 	{
-	  NSNumber	*p = [NSNumber numberWithInt: 0700];
-
-	  attr = [NSDictionary dictionaryWithObject: p
-					     forKey: NSFilePosixPermissions];
 	  if ([manager createDirectoryAtPath: tempDirName
-				  attributes: attr] == NO)
+                 withIntermediateDirectories: YES
+				  attributes: attr
+                                       error: NULL] == NO)
 	    {
 	      NSWarnFLog(@"Attempt to create a secure temporary"
 	        @" directory (%@) failed.", tempDirName);
 	      return nil;
 	    }
 	}
+      else
+        {
+          [manager changeFileAttributes: attr atPath: tempDirName];
+        }
 
       /*
        * Check that the new directory is really secure.
@@ -2449,3 +2489,14 @@ if (domainMask & mask) \
 
   return paths;
 }
+
+@implementation GSPathUtilities
++ (void) atExit
+{
+  ShutdownPathUtilities();
+}
++ (void) initialize
+{
+  [self registerAtExit];
+}
+@end
