@@ -23,7 +23,7 @@
    Boston, MA 02111 USA.
 
    <title>NSProcessInfo class reference</title>
-   $Date: 2012-01-03 03:31:41 -0700 (Tue, 03 Jan 2012) $ $Revision: 34399 $
+   $Date: 2013-08-22 16:44:54 +0100 (Thu, 22 Aug 2013) $ $Revision: 37003 $
 */
 
 /*************************************************************************
@@ -57,16 +57,11 @@
 #import "common.h"
 
 #include <stdio.h>
-#include <string.h>
 
 #ifdef HAVE_WINDOWS_H
 #  include <windows.h>
 #endif
 
-#ifdef HAVE_STRERROR
-#  include <errno.h>
-#endif
-
 #if	defined(HAVE_SYS_SIGNAL_H)
 #  include	<sys/signal.h>
 #elif	defined(HAVE_SIGNAL_H)
@@ -113,7 +108,6 @@
 #include <crt_externs.h>
 #endif
 
-#import "GNUstepBase/GSConfig.h"
 #import "Foundation/NSArray.h"
 #import "Foundation/NSSet.h"
 #import "Foundation/NSCharacterSet.h"
@@ -763,8 +757,7 @@ static char	**_gnu_noobjc_env = NULL;
   if (_gnu_noobjc_argv == NULL)
     goto malloc_error;
 
-
-  ifp=fopen(proc_file_name,"r");
+  ifp = fopen(proc_file_name,"r");
   //freopen(proc_file_name, "r", ifp);
   if (ifp == NULL)
     {
@@ -786,17 +779,25 @@ static char	**_gnu_noobjc_env = NULL;
 	  argument++;
 	  length = 0;
 	  if (c == EOF) // End of command line
-	    break;
+	    {
+	      _gnu_noobjc_argc = argument;
+	      break;
+	    }
 	}
     }
   fclose(ifp);
-  ifp=fopen(proc_file_name,"r");
+  ifp = fopen(proc_file_name,"r");
   //freopen(proc_file_name, "r", ifp);
   if (ifp == NULL)
     {
-      for (c = 0; c < _gnu_noobjc_argc; c++)
-	free(_gnu_noobjc_argv[c]);
-      free(_gnu_noobjc_argv);
+      if (0 != _gnu_noobjc_argv)
+	{
+	  for (c = 0; c < _gnu_noobjc_argc; c++)
+	    {
+	      free(_gnu_noobjc_argv[c]);
+	    }
+	  free(_gnu_noobjc_argv);
+	}
       goto proc_fs_error;
     }
   argument = 0;
@@ -845,6 +846,7 @@ static char	**_gnu_noobjc_env = NULL;
 #endif /* HAVE_FUNCTION_STRERROR */
   fprintf(stderr, "Your gnustep-base library is compiled for a kernel supporting the /proc filesystem, but it can't access it.\n");
   fprintf(stderr, "You should recompile or change your kernel.\n");
+  free(proc_file_name);
 #ifdef HAVE_PROGRAM_INVOCATION_NAME
   fprintf(stderr, "We try to go on anyway; but the program will ignore any argument which were passed to it.\n");
   _gnu_noobjc_argc = 1;
