@@ -26,7 +26,7 @@
    Boston, MA 02111 USA.
 
    <title>NSRunLoop class reference</title>
-   $Date: 2011-10-14 09:46:13 -0600 (Fri, 14 Oct 2011) $ $Revision: 33993 $
+   $Date: 2013-12-20 13:32:29 +0000 (Fri, 20 Dec 2013) $ $Revision: 37484 $
 */
 
 #import "common.h"
@@ -61,7 +61,6 @@
 #endif
 #include <math.h>
 #include <time.h>
-#include <string.h>		/* for memset() */
 
 
 NSString * const NSDefaultRunLoopMode = @"NSDefaultRunLoopMode";
@@ -106,7 +105,19 @@ static NSDate	*theFuture = nil;
 
 - (void) fire
 {
-  [target performSelector: selector withObject: argument];
+  NS_DURING
+    {
+      [target performSelector: selector withObject: argument];
+    }
+  NS_HANDLER
+    {
+      NSLog(@"*** NSRunLoop ignoring exception '%@' (reason '%@') "
+        @"raised during performSelector... with target %p "
+        @"and selector '%@'",
+        [localException name], [localException reason], target,
+        NSStringFromSelector([target selector]));
+    }
+  NS_ENDHANDLER
 }
 
 - (id) initWithSelector: (SEL)aSelector
@@ -695,23 +706,21 @@ static inline BOOL timerInvalidated(NSTimer *t)
     {
       [self currentRunLoop];
       theFuture = RETAIN([NSDate distantFuture]);
+      [[NSObject leakAt: &theFuture] release];
     }
 }
 
-/**
- * Returns the run loop instance for the current thread.
- */
-+ (NSRunLoop*) currentRunLoop
++ (NSRunLoop*) _runLoopForThread: (NSThread*) aThread
 {
-  GSRunLoopThreadInfo	*info = GSRunLoopInfoForThread(nil);
+  GSRunLoopThreadInfo	*info = GSRunLoopInfoForThread(aThread);
   NSRunLoop             *current = info->loop;
 
-  if (current == nil)
+  if (nil == current)
     {
       current = info->loop = [[self alloc] _init];
       /* If this is the main thread, set up a housekeeping timer.
        */
-      if ([GSCurrentThread() isMainThread] == YES)
+      if (nil != current && [GSCurrentThread() isMainThread] == YES)
         {
           NSAutoreleasePool		*arp = [NSAutoreleasePool new];
           GSRunLoopCtxt	                *context;
@@ -760,6 +769,16 @@ static inline BOOL timerInvalidated(NSTimer *t)
   return current;
 }
 
++ (NSRunLoop*) currentRunLoop
+{
+  return [self _runLoopForThread: nil];
+}
+
++ (NSRunLoop*) mainRunLoop
+{
+  return [self _runLoopForThread: [NSThread mainThread]];
+}
+
 - (id) init
 {
   DESTROY(self);
@@ -1221,18 +1240,6 @@ updateTimer(NSTimer *t, NSDate *d, NSTimeInterval now)
   [arp drain];
 }
 
-/**
- * Calls -limitDateForMode: to determine if a timeout occurs before the
- * specified date, then calls -acceptInputForMode:beforeDate: to run the
- * loop once.<br />
- * The specified date may be nil ... in which case the loop runs
- * until the limit date of the first input or timeout.<br />
- * If the specified date is in the past, this runs the loop once only,
- * to handle any events already available.<br />
- * If there are no input sources or timers in mode, this method
- * returns NO without running the loop (irrespective of the supplied
- * date argument), otherwise returns YES.
- */
 - (BOOL) runMode: (NSString*)mode beforeDate: (NSDate*)date
 {
   NSAutoreleasePool	*arp = [NSAutoreleasePool new];
@@ -1240,7 +1247,13 @@ updateTimer(NSTimer *t, NSDate *d, NSTimeInterval now)
 
   NSAssert(mode != nil, NSInvalidArgumentException);
 
-  /* Find out how long we can wait before first limit date. */
+  /* Process any pending notifications.
+   */
+  GSPrivateCheckTasks(); 
+  GSPrivateNotifyASAP(mode); 
+
+  /* Find out how long we can wait before first limit date.
+   */
   d = [self limitDateForMode: mode];
   if (d == nil)
     {
@@ -1248,8 +1261,7 @@ updateTimer(NSTimer *t, NSDate *d, NSTimeInterval now)
       return NO;
     }
 
-  /*
-   * Use the earlier of the two dates we have.
+  /* Use the earlier of the two dates we have.
    * Retain the date in case the firing of a timer (or some other event)
    * releases it.
    */
@@ -1480,8 +1492,10 @@ updateTimer(NSTimer *t, NSDate *d, NSTimeInterval now)
 	  if (i % 1000 == 0 && i > context->maxPerformers)
 	    {
 	      context->maxPerformers = i;
-	      NSLog(@"WARNING ... there are %u performers scheduled"
-		@" in mode %@ of %@", i, mode, self);
+              NSLog(@"WARNING ... there are %u performers scheduled"
+                @" in mode %@ of %@\n(Latest: [%@ %@])",
+                i, mode, self, NSStringFromClass([target class]),
+                NSStringFromSelector(aSelector));
 	    }
 	}
       RELEASE(item);
