@@ -24,7 +24,7 @@
    Boston, MA 02111 USA.
 
    <title>NSScanner class reference</title>
-   $Date: 2011-10-14 04:57:40 -0600 (Fri, 14 Oct 2011) $ $Revision: 33982 $
+   $Date: 2013-08-22 21:03:02 +0100 (Thu, 22 Aug 2013) $ $Revision: 37010 $
 */
 
 #import "common.h"
@@ -51,7 +51,6 @@
 #import "Foundation/NSScanner.h"
 #import "Foundation/NSException.h"
 #import "Foundation/NSUserDefaults.h"
-#import "GNUstepBase/NSObject+GNUstepBase.h"
 
 #import "GSPrivate.h"
 
@@ -205,7 +204,7 @@ typedef GSString	*ivars;
     {
       _string = [_holder initWithString: aString];
     }
-  if (GSObjCIsKindOf(c, GSUnicodeStringClass) == YES)
+  else if (GSObjCIsKindOf(c, GSUnicodeStringClass) == YES)
     {
       _string = RETAIN(aString);
     }
@@ -724,32 +723,28 @@ typedef GSString	*ivars;
   /* Check for trailing exponent */
   if ((_scanLocation < myLength()) && ((c == 'e') || (c == 'E')))
     {
+      unsigned int	expScanLocation = _scanLocation;
       int expval;
+      
 
       _scanLocation++;
       if ([self _scanInt: &expval])
 	{
-      /* Check for exponent overflow */
-	if (num)
-	  {
-	    if ((exponent > 0) && (expval > (LONG_MAX - exponent)))
-	      exponent = LONG_MAX;
-	    else if ((exponent < 0) && (expval < (LONG_MIN - exponent)))
-	      exponent = LONG_MIN;
-	    else
-	      exponent += expval;
-	  }
+        /* Check for exponent overflow */
+          if (num)
+            {
+              if ((exponent > 0) && (expval > (LONG_MAX - exponent)))
+                exponent = LONG_MAX;
+              else if ((exponent < 0) && (expval < (LONG_MIN - exponent)))
+                exponent = LONG_MIN;
+              else
+                exponent += expval;
+            }
 	}
       else
 	{
-#ifdef _ACCEPT_BAD_EXPONENTS_
 	  /* Numbers like 1.23eFOO are accepted (as 1.23). */
 	  _scanLocation = expScanLocation;
-#else
-	  /* Numbers like 1.23eFOO are rejected. */
-	  _scanLocation = saveScanLocation;
-	  return NO;
-#endif
 	}
     }
   if (value)
@@ -1219,8 +1214,15 @@ GSScanInt(unichar *buf, unsigned length, int *result)
   return YES;
 }
 
+/* Table of binary powers of 10 represented by bits in a byte.
+ * Used to convert decimal integer exponents to doubles.
+ */
+static double powersOf10[] = {
+  1.0e1, 1.0e2, 1.0e4, 1.0e8, 1.0e16, 1.0e32, 1.0e64, 1.0e128, 1.0e256
+};
+
 /**
- * Scan in a double value in the standard locale ('.' as decimal poNSInteger).<br />
+ * Scan in a double value in the standard locale ('.' as decimal point).<br />
  * Return YES on success, NO on failure.<br />
  * The value pointed to by result is unmodified on failure.<br />
  * No value is returned in result if it is a null pointer.
@@ -1229,106 +1231,196 @@ BOOL
 GSScanDouble(unichar *buf, unsigned length, double *result)
 {
   unichar	c = 0;
-  double	num = 0.0;
-  long int	exponent = 0;
-  BOOL		negative = NO;
-  BOOL		got_dot = NO;
-  BOOL		got_digit = NO;
+  char          mantissa[20];
+  const char    *ptr;
+  double        *d;
+  double        value;
+  double        e;
+  int	        exponent = 0;
+  BOOL	        negativeMantissa = NO;
+  BOOL		negativeExponent = NO;
   unsigned	pos = 0;
+  int           mantissaLength;
+  int           dotPos = -1;
+  int           hi = 0;
+  int           lo = 0;
 
   /* Skip whitespace */
-  while (pos < length && isspace((NSInteger)buf[pos]))
+  while (pos < length && isspace((int)buf[pos]))
     {
       pos++;
     }
-
-  /* Check for sign */
-  if (pos < length)
-    {
-      switch (buf[pos])
-	{
-	  case '+':
-	    pos++;
-	    break;
-	  case '-':
-	    negative = YES;
-	    pos++;
-	    break;
-	}
-    }
-
-    /* Process number */
-  while (pos < length)
-    {
-      c = buf[pos];
-      if ((c >= '0') && (c <= '9'))
-	{
-	  /* Ensure that the number being accumulated will not overflow. */
-	  if (num >= (DBL_MAX / 10.000000001))
-	    {
-	      ++exponent;
-	    }
-	  else
-	    {
-	      num = (num * 10.0) + (c - '0');
-	      got_digit = YES;
-	    }
-            /* Keep track of the number of digits after the decimal point.
-	       If we just divided by 10 here, we would lose precision. */
-	  if (got_dot)
-	    {
-	      --exponent;
-	    }
-        }
-      else if (!got_dot && (c == '.'))
-	{
-	  /* Note that we have found the decimal point. */
-	  got_dot = YES;
-        }
-      else
-	{
-	  /* Any other character terminates the number. */
-	  break;
-        }
-      pos++;
-    }
-  if (!got_digit)
+  if (pos >= length)
     {
       return NO;
     }
 
-  /* Check for trailing exponent */
-  if ((pos < length) && ((c == 'e') || (c == 'E')))
+  /* Check for sign */
+  switch (buf[pos])
     {
-      int expval;
+      case '+':
+	pos++;
+	break;
+      case '-':
+	negativeMantissa = YES;
+	pos++;
+	break;
+    }
+  if (pos >= length)
+    {
+      return NO;
+    }
 
-      pos++;
-      if (GSScanInt(&buf[pos], length - pos, &expval) == YES)
-	{
-      /* Check for exponent overflow */
-	if (num)
-	  {
-	    if ((exponent > 0) && (expval > (LONG_MAX - exponent)))
-	      exponent = LONG_MAX;
-	    else if ((exponent < 0) && (expval < (LONG_MIN - exponent)))
-	      exponent = LONG_MIN;
-	    else
-	      exponent += expval;
-	  }
-	}
+  /* Scan the mantissa ... at most 18 digits and a decimal point.
+   */
+  for (mantissaLength = 0; pos < length && mantissaLength < 19; pos++)
+    {
+      mantissa[mantissaLength] = c = buf[pos];
+      if (!isdigit(c))
+        {
+          if ('.' != c || dotPos >= 0)
+            {
+              break;    // End of mantissa
+            }
+          dotPos = mantissaLength;
+        }
       else
 	{
-	  return NO;
+          mantissaLength++;
 	}
     }
-  if (result)
+  if (0 == mantissaLength)
     {
-      if (num && exponent)
-	num *= pow(10.0, (double) exponent);
-      if (negative)
-	*result = -num;
+      return NO;        // No mantissa ... not a double
+    }
+  if (mantissaLength > 18)
+    {
+      /* Mantissa too long ... ignore excess.
+       */
+      if (19 == mantissaLength && (dotPos < 0 || dotPos >= mantissaLength))
+        {
+          mantissaLength--;
+        }
       else
-	*result = num;
+        {
+          mantissaLength = 19;
+        }
+    }
+  if (dotPos < 0)
+    {
+      dotPos = mantissaLength;
+    }
+  dotPos -= mantissaLength;      // Exponent offset for decimal point
+
+  /* Convert mantissa characters to a double value
+   */
+  for (ptr = mantissa; mantissaLength > 9; mantissaLength -= 1)
+    {
+      c = *ptr;
+      ptr += 1;
+      hi = hi * 10 + (c - '0');
+    }
+  for (; mantissaLength > 0; mantissaLength -= 1)
+    {
+      c = *ptr;
+      ptr += 1;
+      lo = lo * 10 + (c - '0');
+    }
+  value = (1.0e9 * hi) + lo;
+
+  /* Scan the exponent (if any)
+   */
+  if (pos < length && ('E' == (c = buf[pos]) || 'e' == c))
+    {
+      if (++pos >= length)
+        {
+          return NO;    // Missing exponent
+        }
+      c = buf[pos];
+      if ('-' == c)
+        {
+          negativeExponent = YES;
+          if (++pos >= length)
+            {
+              return NO;    // Missing exponent
+            }
+          c = buf[pos];
+        }
+      else if ('+' == c)
+        {
+          if (++pos >= length)
+            {
+              return NO;    // Missing exponent
+            }
+          c = buf[pos];
+        }
+      while (isdigit(c))
+        {
+          exponent = exponent * 10 + (c - '0');
+          if (++pos >= length)
+            {
+              break;
+            }
+          c = buf[pos];
+        }
+    }
+
+  /* Add in the amount to shift the exponent depending on the position
+   * of the decimal point in the mantissa and check the adjusted sign
+   * of the exponent.
+   */
+  if (YES == negativeExponent)
+    {
+      exponent = dotPos - exponent;
+    }
+  else
+    {
+      exponent = dotPos + exponent;
+    }
+  if (exponent < 0)
+    {
+      negativeExponent = YES;
+      exponent = -exponent;
+    }
+  else
+    {
+      negativeExponent = NO;
+    }
+  if (exponent > 511)
+    {
+      return NO;        // Maximum exponent exceeded
+    }
+
+  /* Convert the exponent to a double then apply it to the value from
+   * the mantissa.
+   */
+  e = 1.0;
+  for (d = powersOf10; exponent != 0; exponent >>= 1, d += 1)
+    {
+      if (exponent & 1)
+        {
+          e *= *d;
+        }
+    }
+  if (YES == negativeExponent)
+    {
+      value /= e;
+    }
+  else
+    {
+      value *= e;
+    }
+
+  if (0 != result)
+    {
+      if (YES == negativeMantissa)
+        {
+          *result = -value;
+        }
+      else
+        {
+          *result = value;
+        }
     }
   return YES;
 }
