@@ -22,7 +22,7 @@
    Boston, MA 02111 USA.
 
    <title>NSSet class reference</title>
-   $Date: 2012-01-01 00:38:53 -0700 (Sun, 01 Jan 2012) $ $Revision: 34374 $
+   $Date: 2013-08-22 16:44:54 +0100 (Thu, 22 Aug 2013) $ $Revision: 37003 $
    */
 
 #import "common.h"
@@ -37,8 +37,8 @@
 // For private method _decodeArrayOfObjectsForKey:
 #import "Foundation/NSKeyedArchiver.h"
 #import "GSPrivate.h"
-#import "GNUstepBase/NSObject+GNUstepBase.h"
 #import "GSFastEnumeration.h"
+#import "GSDispatch.h"
 
 @class	GSSet;
 @interface GSSet : NSObject	// Help the compiler
@@ -260,7 +260,7 @@ static Class NSMutableSet_concrete_class;
         {
 	  unsigned	i;
 	  GS_BEGINIDBUF(objs, count);
-	
+
 	  for (i = 0; i < count; i++)
 	    {
 	      [aCoder decodeValueOfObjCType: @encode(id) at: &objs[i]];
@@ -663,7 +663,7 @@ static Class NSMutableSet_concrete_class;
                 {
                   NSEnumerator  *e = [self objectEnumerator];
                   id            o;
-                  
+
                   while ((o = [e nextObject]) != nil)
                     {
                       d += [[o valueForKeyPath: rem] doubleValue];
@@ -678,7 +678,7 @@ static Class NSMutableSet_concrete_class;
                 {
                   NSEnumerator  *e = [self objectEnumerator];
                   id            o;
-                  
+
                   while ((o = [e nextObject]) != nil)
                     {
                       o = [o valueForKeyPath: rem];
@@ -696,7 +696,7 @@ static Class NSMutableSet_concrete_class;
                 {
                   NSEnumerator  *e = [self objectEnumerator];
                   id            o;
-                  
+
                   while ((o = [e nextObject]) != nil)
                     {
                       o = [o valueForKeyPath: rem];
@@ -716,7 +716,7 @@ static Class NSMutableSet_concrete_class;
                 {
                   NSEnumerator  *e = [self objectEnumerator];
                   id            o;
-                  
+
                   while ((o = [e nextObject]) != nil)
                     {
                       d += [[o valueForKeyPath: rem] doubleValue];
@@ -730,7 +730,7 @@ static Class NSMutableSet_concrete_class;
                 {
                   NSEnumerator  *e = [self objectEnumerator];
                   id            o;
-                  
+
                   result = [NSMutableSet set];
                   while ((o = [e nextObject]) != nil)
                     {
@@ -750,7 +750,7 @@ static Class NSMutableSet_concrete_class;
                 {
                   NSEnumerator  *e = [self objectEnumerator];
                   id            o;
-                  
+
                   result = [NSMutableSet set];
                   while ((o = [e nextObject]) != nil)
                     {
@@ -770,7 +770,7 @@ static Class NSMutableSet_concrete_class;
                 {
                   NSEnumerator  *e = [self objectEnumerator];
                   id            o;
-                  
+
                   result = [NSMutableSet set];
                   while ((o = [e nextObject]) != nil)
                     {
@@ -790,7 +790,7 @@ static Class NSMutableSet_concrete_class;
                 {
                   NSEnumerator  *e = [self objectEnumerator];
                   id            o;
-                  
+
                   result = [GSMutableArray array];
                   while ((o = [e nextObject]) != nil)
                     {
@@ -810,7 +810,7 @@ static Class NSMutableSet_concrete_class;
                 {
                   NSEnumerator  *e = [self objectEnumerator];
                   id            o;
-                  
+
                   result = [GSMutableArray array];
                   while ((o = [e nextObject]) != nil)
                     {
@@ -830,7 +830,7 @@ static Class NSMutableSet_concrete_class;
                 {
                   NSEnumerator  *e = [self objectEnumerator];
                   id            o;
-                  
+
                   result = [GSMutableArray array];
                   while ((o = [e nextObject]) != nil)
                     {
@@ -866,18 +866,52 @@ static Class NSMutableSet_concrete_class;
 - (void) enumerateObjectsWithOptions: (NSEnumerationOptions)opts
                           usingBlock: (GSSetEnumeratorBlock)aBlock
 {
-  BOOL shouldStop = NO;
+  BLOCK_SCOPE BOOL shouldStop = NO;
   id<NSFastEnumeration> enumerator = self;
-  
+
+  GS_DISPATCH_CREATE_QUEUE_AND_GROUP_FOR_ENUMERATION(enumQueue, opts)
   FOR_IN (id, obj, enumerator)
   {
-    CALL_BLOCK(aBlock, obj, &shouldStop);
+    GS_DISPATCH_SUBMIT_BLOCK(enumQueueGroup,enumQueue, if (shouldStop) {return;}, return;, aBlock, obj, &shouldStop);
     if (shouldStop)
       {
-	return;
+	break;
       }
   }
   END_FOR_IN(enumerator)
+  GS_DISPATCH_TEARDOWN_QUEUE_AND_GROUP_FOR_ENUMERATION(enumQueue, opts)
+}
+
+- (NSSet *) objectsPassingTest: (GSSetFilterBlock)aBlock
+{
+  return [self objectsWithOptions: 0 passingTest: aBlock];
+}
+
+- (NSSet *) objectsWithOptions: (NSEnumerationOptions)opts
+                   passingTest: (GSSetFilterBlock)aBlock
+{
+  BOOL                  shouldStop = NO;
+  id<NSFastEnumeration> enumerator = self;
+  NSMutableSet          *resultSet;
+
+  resultSet = [NSMutableSet setWithCapacity: [self count]];
+    
+  FOR_IN (id, obj, enumerator)
+    {
+      BOOL include = CALL_BLOCK(aBlock, obj, &shouldStop);
+
+      if (include)
+        {
+          [resultSet addObject:obj];
+        }
+      if (shouldStop)
+        {
+          break;
+        }
+    }
+  END_FOR_IN(enumerator)
+    
+  return [resultSet makeImmutableCopyOnFail: NO];
 }
 
 /** Return a set formed by adding anObject to the receiver.
@@ -922,7 +956,7 @@ static Class NSMutableSet_concrete_class;
   return [s autorelease];
 }
 
-- (NSUInteger) countByEnumeratingWithState: (NSFastEnumerationState*)state 	
+- (NSUInteger) countByEnumeratingWithState: (NSFastEnumerationState*)state
                                    objects: (id*)stackbuf
                                      count: (NSUInteger)len
 {
