@@ -21,9 +21,13 @@
    Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA 02111 USA.
 
-   $Date: 2011-03-09 05:40:50 -0700 (Wed, 09 Mar 2011) $ $Revision: 32517 $
+   $Date: 2013-08-22 16:44:54 +0100 (Thu, 22 Aug 2013) $ $Revision: 37003 $
    */
 
+/* define to get system-v functions including inet_aton()
+ */
+#define _SVID_SOURCE    1
+
 #import "common.h"
 #define	EXPOSE_NSSocketPortNameServer_IVARS	1
 #import "Foundation/NSData.h"
@@ -557,10 +561,13 @@ typedef enum {
   if (self == [NSSocketPortNameServer class])
     {
       serverLock = [NSRecursiveLock new];
+      [[NSObject leakAt: &serverLock] release];
       modes = [[NSArray alloc] initWithObjects: &mode count: 1];
+      [[NSObject leakAt: &modes] release];
 #ifdef	GDOMAP_PORT_OVERRIDE
       serverPort = RETAIN([NSString stringWithUTF8String:
 	make_gdomap_port(GDOMAP_PORT_OVERRIDE)]);
+      [[NSObject leakAt: &serverPort] release];
 #endif
       portClass = [NSSocketPort class];
     }
