@@ -22,7 +22,7 @@
    Boston, MA 02111 USA.
 
    <title>NSTask class reference</title>
-   $Date: 2011-12-15 02:42:39 -0700 (Thu, 15 Dec 2011) $ $Revision: 34290 $
+   $Date: 2013-12-20 13:32:29 +0000 (Fri, 20 Dec 2013) $ $Revision: 37484 $
    */
 
 #import "common.h"
@@ -45,10 +45,8 @@
 #import "Foundation/NSTimer.h"
 #import "Foundation/NSLock.h"
 #import "GNUstepBase/NSString+GNUstepBase.h"
-#import "GNUstepBase/NSObject+GNUstepBase.h"
 #import "GSPrivate.h"
 
-#include <string.h>
 #include <sys/types.h>
 #if defined(__FreeBSD__) || defined(__OpenBSD__)
 #include <fcntl.h>
@@ -260,6 +258,7 @@ pty_slave(const char* name)
       if (tasksLock == nil)
         {
           tasksLock = [NSRecursiveLock new];
+          [[NSObject leakAt: &tasksLock] release];
 	  /* The activeTasks map contains the NSTask objects corresponding
 	   * to running subtasks, and retains them until the subprocess
 	   * actually terminates.
@@ -284,6 +283,7 @@ pty_slave(const char* name)
 	   */
           activeTasks = NSCreateMapTable(NSIntegerMapKeyCallBacks,
                 NSObjectMapValueCallBacks, 0);
+          [[NSObject leakAt: &activeTasks] release];
         }
       [gnustep_global_lock unlock];
 
@@ -854,14 +854,15 @@ pty_slave(const char* name)
  */
 - (void) waitUntilExit
 {
+  CREATE_AUTORELEASE_POOL(arp);
+  NSRunLoop     *loop = [NSRunLoop currentRunLoop];
   NSTimer	*timer = nil;
+  NSDate	*limit = nil;
 
+  IF_NO_GC([[self retain] autorelease];)
   while ([self isRunning])
     {
-      NSDate	*limit;
-
-      /*
-       *	Poll at 0.1 second intervals.
+      /* Poll at 0.1 second intervals.
        */
       limit = [[NSDate alloc] initWithTimeIntervalSinceNow: 0.1];
       if (timer == nil)
@@ -872,11 +873,17 @@ pty_slave(const char* name)
 						 userInfo: nil
 						  repeats: YES];
 	}
-      [[NSRunLoop currentRunLoop] runMode: NSDefaultRunLoopMode
-			       beforeDate: limit];
-      RELEASE(limit);
+      [loop runMode: NSDefaultRunLoopMode beforeDate: limit];
+      DESTROY(limit);
     }
   [timer invalidate];
+
+  /* Run loop one last time (with limit date in past) so that any
+   * notification about the task ending is sent immediately.
+   */
+  limit = [NSDate dateWithTimeIntervalSinceNow: 0.0];
+  [loop runMode: NSDefaultRunLoopMode beforeDate: limit];
+  IF_NO_GC([arp release];)
 }
 @end
 
