@@ -25,7 +25,7 @@
    Boston, MA 02111 USA.
 
    <title>NSTimeZone class reference</title>
-   $Date: 2011-07-31 09:31:39 -0600 (Sun, 31 Jul 2011) $ $Revision: 33660 $
+   $Date: 2013-10-29 07:03:29 +0000 (Tue, 29 Oct 2013) $ $Revision: 37312 $
  */
 
 /* Use the system time zones if available. In other cases, use an
@@ -88,7 +88,6 @@
 #define	EXPOSE_NSTimeZone_IVARS	1
 #import "GNUstepBase/GSLock.h"
 #include <stdio.h>
-#include <string.h>
 #include <time.h>
 #import "Foundation/NSArray.h"
 #import "Foundation/NSCoder.h"
@@ -107,8 +106,6 @@
 #import "Foundation/NSTimeZone.h"
 #import "Foundation/NSByteOrder.h"
 #import "Foundation/NSLocale.h"
-#import "GNUstepBase/GSConfig.h"
-#import "GNUstepBase/NSObject+GNUstepBase.h"
 #import "GNUstepBase/NSString+GNUstepBase.h"
 #import "GSPrivate.h"
 
@@ -154,7 +151,7 @@ NSString * const NSSystemTimeZoneDidChangeNotification
 #define LOCAL_TIME_FILE @"localtime"
 
 /* Directory that contains the actual time zones. */
-#define ZONES_DIR @"zones/"
+#define ZONES_DIR @"zones"
 
 /* Many systems have this file */
 #define SYSTEM_TIME_FILE @"/etc/localtime"
@@ -167,6 +164,7 @@ NSString * const NSSystemTimeZoneDidChangeNotification
 #endif
 
 #define BUFFER_SIZE 512
+#define WEEK_MILLISECONDS (7.0*24.0*60.0*60.0*1000.0)
 
 #if GS_USE_ICU == 1
 static inline int
@@ -298,30 +296,6 @@ static NSRecursiveLock *zone_mutex = nil;
 static Class	NSTimeZoneClass;
 static Class	GSPlaceholderTimeZoneClass;
 
-/* Decode the four bytes at PTR as a signed integer in network byte order.
-   Based on code included in the GNU C Library 2.0.3. */
-static inline int
-decode (const void *ptr)
-{
-#if defined(WORDS_BIGENDIAN) && SIZEOF_INT == 4
-#if NEED_WORD_ALIGNMENT
-  int value;
-  memcpy(&value, ptr, sizeof(NSInteger));
-  return value;
-#else
-  return *(const NSInteger*) ptr;
-#endif
-#else /* defined(WORDS_BIGENDIAN) && SIZEOF_INT == 4 */
-  const unsigned char *p = ptr;
-  int result = *p & (1 << (CHAR_BIT - 1)) ? ~0 : 0;
-
-  result = (result << 8) | *p++;
-  result = (result << 8) | *p++;
-  result = (result << 8) | *p++;
-  result = (result << 8) | *p;
-  return result;
-#endif /* defined(WORDS_BIGENDIAN) && SIZEOF_INT == 4 */
-}
 
 /* Return path to a TimeZone directory file */
 static NSString *_time_zone_path(NSString *subpath, NSString *type)
@@ -761,8 +735,8 @@ static NSMapTable	*absolutes = 0;
 	       * Should never happen now we round to the minute
 	       * for MacOS-X compatibnility.
 	       */
-	      name = [[NSString alloc] initWithFormat: @"NSAbsoluteTimeZone:%d",
-		anOffset];
+	      name = [[NSString alloc]
+		initWithFormat: @"NSAbsoluteTimeZone:%"PRIdPTR, anOffset];
 	    }
 	}
       else
@@ -1359,6 +1333,12 @@ static NSMapTable	*absolutes = 0;
       zone_mutex = [GSLazyRecursiveLock new];
       [[NSObject leakAt: (id*)&zone_mutex] release];
 
+      [[NSObject leakAt: (id*)&defaultTimeZone] release];
+      [[NSObject leakAt: (id*)&systemTimeZone] release];
+      [[NSObject leakAt: (id*)&abbreviationDictionary] release];
+      [[NSObject leakAt: (id*)&abbreviationMap] release];
+      [[NSObject leakAt: (id*)&absolutes] release];
+
       [[NSNotificationCenter defaultCenter] addObserver: self
         selector: @selector(_notified:)
         name: NSUserDefaultsDidChangeNotification
@@ -2056,6 +2036,11 @@ localZoneString, [zone name], sign, s/3600, (s/60)%60);
   [aCoder encodeObject: [self name]];
 }
 
+- (NSUInteger) hash
+{
+  return [[self name] hash];
+}
+
 - (id) init
 {
   return [self initWithName: @"NSLocalTimeZone" data: nil];
@@ -2125,7 +2110,7 @@ localZoneString, [zone name], sign, s/3600, (s/60)%60);
 }
 
 /**
- * Returns TRUE if the time zones have the same name.
+ * Returns YES if the time zones have the same name.
  */
 - (BOOL) isEqualToTimeZone: (NSTimeZone*)aTimeZone
 {
@@ -2232,7 +2217,73 @@ localZoneString, [zone name], sign, s/3600, (s/60)%60);
 
 - (NSDate *) nextDaylightSavingTimeTransitionAfterDate: (NSDate *)aDate
 {
+#if GS_USE_ICU == 1
+  /* ICU doesn't provide transition information per se.
+   * The canonical method of retrieving this piece of information is to
+   * use binary search.
+   */
+  
+  int32_t originalOffset, currentOffset;
+  UCalendar *cal;
+  UErrorCode err = U_ZERO_ERROR;
+  UDate currentTime;
+  int i;
+  NSDate* result = nil;
+  
+  cal = ICUCalendarSetup (self, nil);
+  if (cal == NULL)
+    return nil;
+  
+  currentTime = [aDate timeIntervalSince1970] * 1000.0;
+  ucal_setMillis (cal, currentTime, &err);
+  originalOffset = ucal_get (cal, UCAL_DST_OFFSET, &err);
+  if (U_FAILURE(err))
+    return nil;
+
+  /* First try to find the next transition by adding a week at a time */
+  /* Avoid ending in an infinite loop in case there is no transition at all */
+
+  for (i = 0; i < 53; i++)
+    {
+      /* Add a single week */
+      currentTime += WEEK_MILLISECONDS;
+
+	  ucal_setMillis (cal, currentTime, &err);
+	  if (U_FAILURE(err))
+        break;
+
+	  currentOffset = ucal_get (cal, UCAL_DST_OFFSET, &err);
+	  if (U_FAILURE(err))
+        break;
+
+	  if (currentOffset != originalOffset)
+        {
+          double interval = WEEK_MILLISECONDS / 2.0;
+          /* Now use bisection to determine the exact moment */
+
+		  while (interval >= 1.0)
+            {
+              ucal_setMillis (cal, currentTime - interval, &err);
+
+              currentOffset = ucal_get (cal, UCAL_DST_OFFSET, &err);
+
+              if (currentOffset != originalOffset)
+                currentTime -= interval; /* it is in the lower half */
+
+              interval /= 2.0;
+            }
+
+           result =
+             [NSDate dateWithTimeIntervalSince1970: floor(currentTime/1000.0)];
+        }
+    }
+
+  ucal_close (cal);
+  
+  return result;
+#else
   return nil;   // FIXME;
+#endif
 }
 
 - (NSTimeInterval) daylightSavingTimeOffset
@@ -2289,7 +2340,7 @@ localZoneString, [zone name], sign, s/3600, (s/60)%60);
 
 - (NSString*) description
 {
-  return [NSString stringWithFormat: @"%@(%@, %s%d)", [self name],
+  return [NSString stringWithFormat: @"%@(%@, %s%"PRIdPTR")", [self name],
     [self timeZoneAbbreviation],
     ([self isDaylightSavingTimeZone]? "IS_DST, ": ""),
     [self timeZoneSecondsFromGMT]];
@@ -3043,10 +3094,12 @@ newDetailInZoneForType(GSTimeZone *zone, TypeInfo *type)
       for (i = 0; i < n_types; i++)
 	{
 	  struct ttinfo	*ptr = (struct ttinfo*)(bytes + pos);
+          uint32_t      off;
 
 	  types[i].isdst = (ptr->isdst != 0 ? YES : NO);
 	  types[i].abbr_idx = ptr->abbr_idx;
-	  types[i].offset = decode(ptr->offset);
+          memcpy(&off, ptr->offset, 4);
+	  types[i].offset = GSSwapBigI32ToHost(off);
 	  pos += sizeof(struct ttinfo);
 	}
       abbr = (unsigned char*)(bytes + pos);
