@@ -33,9 +33,9 @@
 #import "Foundation/NSValue.h"
 
 #import "GSPrivate.h"
+#import "GSTLS.h"
 #import "GSURLPrivate.h"
 #import "GNUstepBase/GSMime.h"
-#import "GNUstepBase/NSObject+GNUstepBase.h"
 #import "GNUstepBase/NSString+GNUstepBase.h"
 #import "GNUstepBase/NSURL+GNUstepBase.h"
 
@@ -97,7 +97,9 @@ static NSLock		*pairLock = nil;
        * to the same value.
        */
       pairCache = [NSMutableArray new];
+      [[NSObject leakAt: &pairCache] release];
       pairLock = [NSLock new];
+      [[NSObject leakAt: &pairLock] release];
       /*  Purge expired pairs at intervals.
        */
       [[NSNotificationCenter defaultCenter] addObserver: self
@@ -282,6 +284,8 @@ static NSLock		*pairLock = nil;
 @interface _NSHTTPSURLProtocol : _NSHTTPURLProtocol
 @end
 
+@interface _NSDataURLProtocol : NSURLProtocol
+@end
 
 
 // Internal data storage
@@ -363,13 +367,17 @@ static NSURLProtocol	*placeholder = nil;
       placeholderClass = [NSURLProtocolPlaceholder class];
       placeholder = (NSURLProtocol*)NSAllocateObject(placeholderClass, 0,
 	NSDefaultMallocZone());
+      [[NSObject leakAt: &placeholder] release];
       registered = [NSMutableArray new];
+      [[NSObject leakAt: &registered] release];
       regLock = [NSLock new];
+      [[NSObject leakAt: &regLock] release];
       [self registerClass: [_NSHTTPURLProtocol class]];
       [self registerClass: [_NSHTTPSURLProtocol class]];
       [self registerClass: [_NSFTPURLProtocol class]];
       [self registerClass: [_NSFileURLProtocol class]];
       [self registerClass: [_NSAboutURLProtocol class]];
+      [self registerClass: [_NSDataURLProtocol class]];
     }
 }
 
@@ -536,6 +544,14 @@ static NSURLProtocol	*placeholder = nil;
   return this->request;
 }
 
+/* This method is here so that it's safe to set debug on any NSURLProtocol
+ * even if the concrete subclass doesn't actually support debug logging.
+ */
+- (void) setDebug: (BOOL)flag
+{
+  return;
+}
+
 @end
 
 
@@ -615,6 +631,15 @@ static NSURLProtocol	*placeholder = nil;
   [super dealloc];
 }
 
+- (id) init
+{
+  if (nil != (self = [super init]))
+    {
+      _debug = GSDebugSet(@"NSURLProtocol");
+    }
+  return self;
+}
+
 - (void) setDebug: (BOOL)flag
 {
   _debug = flag;
@@ -656,7 +681,6 @@ static NSURLProtocol	*placeholder = nil;
   _statusCode = 0;	/* No status returned yet.	*/
   _isLoading = YES;
   _complete = NO;
-  _debug = GSDebugSet(@"NSHTTPURLProtocol");
 
   /* Perform a redirect if the path is empty.
    * As per MacOs-X documentation.
@@ -764,10 +788,39 @@ static NSURLProtocol	*placeholder = nil;
 #endif
       if ([[url scheme] isEqualToString: @"https"] == YES)
         {
+          static NSArray        *keys;
+          NSUInteger            count;
+
           [this->input setProperty: NSStreamSocketSecurityLevelNegotiatedSSL
                             forKey: NSStreamSocketSecurityLevelKey];
           [this->output setProperty: NSStreamSocketSecurityLevelNegotiatedSSL
                              forKey: NSStreamSocketSecurityLevelKey];
+          if (nil == keys)
+            {
+              keys = [[NSArray alloc] initWithObjects:
+                GSTLSCAFile,
+                GSTLSCertificateFile,
+                GSTLSCertificateKeyFile,
+                GSTLSCertificateKeyPassword,
+                GSTLSDebug,
+                GSTLSPriority,
+                GSTLSRemoteHosts,
+                GSTLSRevokeFile,
+                GSTLSVerify,
+                nil];
+            }
+          count = [keys count];
+          while (count-- > 0)
+            {
+              NSString      *key = [keys objectAtIndex: count];
+              NSString      *str = [this->request _propertyForKey: key];
+
+              if (nil != str)
+                {
+                  [this->output setProperty: str forKey: key];
+                }
+            }
+          if (_debug) [this->output setProperty: @"YES" forKey: GSTLSDebug];
         }
       [this->input setDelegate: self];
       [this->output setDelegate: self];
@@ -864,10 +917,10 @@ static NSURLProtocol	*placeholder = nil;
       GSMimeDocument	*document = [_parser mimeDocument];
       unsigned		bodyLength;
 
-      if (wasInHeaders == YES && isInHeaders == NO)
+      _complete = [_parser isComplete];
+      if (YES == wasInHeaders && NO == isInHeaders)
         {
 	  GSMimeHeader		*info;
-	  NSString		*enc;
 	  int			len = -1;
 	  NSString		*ct;
 	  NSString		*st;
@@ -900,11 +953,15 @@ static NSURLProtocol	*placeholder = nil;
 	    }
 
 	  s = [info objectForKey: NSHTTPPropertyStatusReasonKey];
+
+/* Should use this?
+	  NSString		*enc;
 	  enc = [[document headerNamed: @"content-transfer-encoding"] value];
 	  if (enc == nil)
 	    {
 	      enc = [[document headerNamed: @"transfer-encoding"] value];
 	    }
+*/
 
 	  info = [document headerNamed: @"content-type"];
 	  ct = [document contentType];
@@ -930,11 +987,7 @@ static NSURLProtocol	*placeholder = nil;
 	    {
 	      _complete = YES;	// No body expected.
 	    }
-	  else if ([enc isEqualToString: @"chunked"] == YES)	
-	    {
-	      _complete = NO;	// Read chunked body data
-	    }
-	  if (_complete == NO && [d length] == 0)
+	  else if (_complete == NO && [d length] == 0)
 	    {
 	      _complete = YES;	// Had EOF ... terminate
 	    }
@@ -976,6 +1029,26 @@ static NSURLProtocol	*placeholder = nil;
 	    {
 	      NSURLCacheStoragePolicy policy;
 
+	      /* Get cookies from the response and accept them into
+	       * shared storage if policy permits
+	       */
+	      if ([this->request HTTPShouldHandleCookies] == YES
+		&& [_response isKindOfClass: [NSHTTPURLResponse class]] == YES)
+		{
+		  NSDictionary	*hdrs;
+		  NSArray	*cookies;
+		  NSURL		*url;
+
+		  url = [_response URL];
+		  hdrs = [_response allHeaderFields];
+		  cookies = [NSHTTPCookie cookiesWithResponseHeaderFields: hdrs
+								   forURL: url];
+		  [[NSHTTPCookieStorage sharedHTTPCookieStorage]
+		    setCookies: cookies
+		    forURL: url
+		    mainDocumentURL: [this->request mainDocumentURL]];
+		}
+
 	      /* Tell the client that we have a response and how
 	       * it should be cached.
 	       */
@@ -1496,6 +1569,16 @@ static NSURLProtocol	*placeholder = nil;
 				  _writeOffset = 0;
 				}
 			    }
+                          else if ([this->output streamStatus]
+                            == NSStreamStatusWriting)
+                            {
+                              /* Couldn't write it all now, save and try
+                               * again later.
+                               */
+                              _writeData = [[NSData alloc] initWithBytes:
+                                buffer length: len];
+                              _writeOffset = 0;
+                            }
 			}
 		      else
 		        {
@@ -1754,6 +1837,93 @@ static NSURLProtocol	*placeholder = nil;
 
 @end
 
+@implementation _NSDataURLProtocol
+
++ (BOOL) canInitWithRequest: (NSURLRequest*)request
+{
+  return [[[request URL] scheme] isEqualToString: @"data"];
+}
+
++ (NSURLRequest*) canonicalRequestForRequest: (NSURLRequest*)request
+{
+  return request;
+}
+
+- (void) startLoading
+{
+  NSURLResponse *r;
+  NSString      *mime = @"text/plain";
+  NSString      *encoding = @"US-ASCII";
+  NSData        *data;
+  NSString      *spec = [[this->request URL] resourceSpecifier];
+  NSRange       comma = [spec rangeOfString:@","];
+  NSEnumerator  *types;
+  NSString      *type;
+  BOOL          base64 = NO;
+
+  if (comma.location == NSNotFound)
+    {
+      NSDictionary      *ui;
+      NSError           *error;
+
+      ui = [NSDictionary dictionaryWithObjectsAndKeys:
+        [this->request URL], @"URL",
+        [[this->request URL] path], @"path",
+        nil];
+      error = [NSError errorWithDomain: @"can't load data"
+                                  code: 0
+                              userInfo: ui];
+      [this->client URLProtocol: self didFailWithError: error];
+      return;
+    }
+  types = [[[spec substringToIndex: comma.location]
+    componentsSeparatedByString: @";"] objectEnumerator];
+  while (nil != (type = [types nextObject]))
+    {
+      if ([type isEqualToString: @"base64"])
+	{
+	  base64 = YES;
+	}
+      else if ([type hasPrefix: @"charset="])
+	{
+	  encoding = [type substringFromIndex: 8];
+	}
+      else if ([type length] > 0)
+	{
+	  mime = type;
+	}
+    }
+  spec = [spec substringFromIndex: comma.location + 1];
+  if (YES == base64)
+    {
+      data = [GSMimeDocument decodeBase64:
+        [spec dataUsingEncoding: NSUTF8StringEncoding]];
+    }
+  else
+    {
+      data = [[spec stringByReplacingPercentEscapesUsingEncoding:
+        NSUTF8StringEncoding] dataUsingEncoding: NSUTF8StringEncoding];
+    }
+  r = [[NSURLResponse alloc] initWithURL: [this->request URL]
+    MIMEType: mime
+    expectedContentLength: [data length]
+    textEncodingName: encoding];
+
+  [this->client URLProtocol: self
+         didReceiveResponse: r 
+	 cacheStoragePolicy: NSURLRequestUseProtocolCachePolicy];
+  [this->client URLProtocol: self didLoadData: data];
+  [this->client URLProtocolDidFinishLoading: self];
+  RELEASE(r);
+}
+
+- (void) stopLoading
+{
+  return;
+}
+
+@end
+
 @implementation _NSAboutURLProtocol
 
 + (BOOL) canInitWithRequest: (NSURLRequest*)request
