@@ -21,7 +21,7 @@
    Boston, MA 02111 USA.
 
    <title>NSUndoManager class reference</title>
-   $Date: 2011-02-14 23:25:54 -0700 (Mon, 14 Feb 2011) $ $Revision: 32173 $
+   $Date: 2013-11-06 01:52:54 +0000 (Wed, 06 Nov 2013) $ $Revision: 37365 $
 */
 
 #import "common.h"
@@ -695,7 +695,7 @@
 	  return [NSString stringWithFormat: @"%@ %@", _(@"Redo"), actionName];
 	}
     }
-  return actionName;
+  return _(@"Redo");
 }
 
 /**
@@ -944,7 +944,15 @@
     {
       return nil;
     }
-  return [[_undoStack lastObject] actionName];
+
+  if (_group != nil)
+    {
+      return [_group actionName];
+    }
+  else
+    {
+      return [[_undoStack lastObject] actionName];
+    }
 }
 
 /**
@@ -975,7 +983,7 @@
 	  return [NSString stringWithFormat: @"%@ %@", _(@"Undo"), actionName];
 	}
     }
-  return actionName;
+  return _(@"Undo");
 }
 
 /**
