@@ -24,7 +24,7 @@
    Boston, MA 02111 USA.
 
    <title>NSUserDefaults class reference</title>
-   $Date: 2012-01-29 04:10:05 -0700 (Sun, 29 Jan 2012) $ $Revision: 34654 $
+   $Date: 2013-11-20 06:12:29 +0000 (Wed, 20 Nov 2013) $ $Revision: 37385 $
 */
 
 #import "common.h"
@@ -54,7 +54,6 @@
 #import "GNUstepBase/GSLocale.h"
 #import "GNUstepBase/GSLock.h"
 #import "GNUstepBase/NSProcessInfo+GNUstepBase.h"
-#import "GNUstepBase/NSObject+GNUstepBase.h"
 #import "GNUstepBase/NSString+GNUstepBase.h"
 
 #if	defined(__MINGW__)
@@ -101,6 +100,7 @@ static BOOL		hasSharedDefaults = NO;
 /*
  * Caching some defaults.
  */
+static BOOL     parsingArguments = NO;
 static BOOL	flags[GSUserDefaultMaxFlag] = { 0 };
 
 /* An instance of the GSPersistentDomain class is used to encapsulate
@@ -163,7 +163,10 @@ lockPath(NSString *defaultsDatabase, BOOL verbose)
 
   if ([mgr fileExistsAtPath: path isDirectory: &isDir] == NO)
     {
-      if ([mgr createDirectoryAtPath: path attributes: attr] == NO)
+      if ([mgr createDirectoryAtPath: path
+         withIntermediateDirectories: YES
+                          attributes: attr
+                               error: NULL] == NO)
 	{
 	  if (verbose)
 	    NSLog(@"Defaults path '%@' does not exist - failed to create it.",
@@ -186,7 +189,10 @@ lockPath(NSString *defaultsDatabase, BOOL verbose)
   path = [path stringByAppendingPathComponent: @".lck"];
   if ([mgr fileExistsAtPath: path isDirectory: &isDir] == NO)
     {
-      if ([mgr createDirectoryAtPath: path attributes: attr] == NO)
+      if ([mgr createDirectoryAtPath: path
+         withIntermediateDirectories: YES
+                          attributes: attr
+                               error: NULL] == NO)
 	{
 	  if (verbose)
 	    NSLog(@"Defaults path '%@' does not exist - failed to create it.",
@@ -353,19 +359,16 @@ systemLanguages()
   }	
   
   // If LANGUAGES did not yield any languages, try LC_MESSAGES
-#ifdef HAVE_LOCALE_H
-#ifdef LC_MESSAGES
+
   if ([names count] == 0)
     {
-      NSString *locale = GSSetLocale(LC_MESSAGES, nil);
+      NSString *locale = GSDefaultLanguageLocale();
 
       if (locale != nil)
 	{
 	  [names addObjectsFromArray: GSLanguagesFromLocale(locale)];
 	}
     }
-#endif
-#endif
 
   return names;
 }
@@ -540,11 +543,9 @@ newLanguages(NSArray *oldNames)
 
 + (void) atExit
 {
-  id	tmp;
-
-  tmp = sharedDefaults;
-  sharedDefaults = nil;
-  [tmp release];
+  DESTROY(sharedDefaults);
+  DESTROY(processName);
+  DESTROY(classLock);
 }
 
 + (void) initialize
@@ -994,13 +995,8 @@ newLanguages(NSArray *oldNames)
 	         * precisely because it is the currently set locale in the
 	         * C library.
 	         */
-	        NSString	*locale = nil;
+	        NSString	*locale = GSDefaultLanguageLocale();
 
-#ifdef HAVE_LOCALE_H
-#ifdef LC_MESSAGES
-	        locale = GSSetLocale(LC_MESSAGES, nil);
-#endif
-#endif
 	        if (locale != nil)
 	          {
 		    NSString	*i18n = GSLanguageFromLocale(locale);
@@ -1222,7 +1218,7 @@ newLanguages(NSArray *oldNames)
 
 - (void) addSuiteNamed: (NSString*)aName
 {
-  unsigned	index;
+  NSUInteger	index;
 
   if (aName == nil)
     {
@@ -1775,7 +1771,7 @@ static BOOL isPlistObject(id o)
   NSDate		*saved;
   BOOL			wasLocked;
   BOOL			result = YES;
-  BOOL			haveNewDomain = NO;
+  BOOL			haveChange = NO;
 
   [_lock lock];
   saved = _lastSync;
@@ -1808,42 +1804,46 @@ static BOOL isPlistObject(id o)
 	    {
 	      NSEnumerator		*enumerator;
 	      NSString			*domainName;
-	      NSFileManager		*mgr;
 
-	      haveNewDomain = [self _readDefaults];
-	      if (YES == haveNewDomain)
+	      haveChange = [self _readDefaults];
+	      if (YES == haveChange)
 		{
 		  DESTROY(_dictionaryRep);
 		}
 
-	      mgr = [NSFileManager defaultManager];
+	      if (_changedDomains != nil)
+                {
+                  haveChange = YES;
 
-	      if (_changedDomains != nil && NO == [self _readOnly])
-		{
-		  GSPersistentDomain	*domain;
+                  if (NO == [self _readOnly])
+                    {
+                      GSPersistentDomain	*domain;
+                      NSFileManager		*mgr;
 
-		  enumerator = [_changedDomains objectEnumerator];
-		  DESTROY(_changedDomains);	// Retained by enumerator.
-		  while ((domainName = [enumerator nextObject]) != nil)
-		    {
-		      domain = [_persDomains objectForKey: domainName];
-		      if (domain != nil)	// Domain was added or changed
-			{
-			  [domain synchronize];
-			}
-		      else			// Domain was removed
-			{
-			  NSString	*path;
+                      mgr = [NSFileManager defaultManager];
+                      enumerator = [_changedDomains objectEnumerator];
+                      DESTROY(_changedDomains);	// Retained by enumerator.
+                      while ((domainName = [enumerator nextObject]) != nil)
+                        {
+                          domain = [_persDomains objectForKey: domainName];
+                          if (domain != nil)	// Domain was added or changed
+                            {
+                              [domain synchronize];
+                            }
+                          else			// Domain was removed
+                            {
+                              NSString	*path;
 
-			  path = [[_defaultsDatabase
-			    stringByAppendingPathComponent: domainName]
-			    stringByAppendingPathExtension: @"plist"];
-			  [mgr removeFileAtPath: path handler: nil];
-			}
-		    }
-		}
+                              path = [[_defaultsDatabase
+                                stringByAppendingPathComponent: domainName]
+                                stringByAppendingPathExtension: @"plist"];
+                              [mgr removeFileAtPath: path handler: nil];
+                            }
+                        }
+                    }
+                }
 
-	      if (YES == haveNewDomain)
+	      if (YES == haveChange)
 		{
 		  updateCache(self);
 		}
@@ -1894,7 +1894,7 @@ static BOOL isPlistObject(id o)
       [self _changePersistentDomain: NSGlobalDomain];
     }
   [_lock unlock];
-  if (YES == haveNewDomain)
+  if (YES == haveChange)
     {
       [[NSNotificationCenter defaultCenter]
 	postNotificationName: NSUserDefaultsDidChangeNotification
@@ -2109,7 +2109,7 @@ static BOOL isPlistObject(id o)
 BOOL
 GSPrivateDefaultsFlag(GSUserDefaultFlagType type)
 {
-  if (sharedDefaults == nil)
+  if (nil == sharedDefaults && NO == parsingArguments)
     {
       [NSUserDefaults standardUserDefaults];
     }
@@ -2163,6 +2163,12 @@ NSDictionary *GSPrivateDefaultLocale()
   id		key, val;
 
   [_lock lock];
+  if (YES == parsingArguments)
+    {
+      [_lock unlock];
+      return nil;       // Prevent recursion
+    }
+  parsingArguments = YES;
   NS_DURING
     {
       args = [[NSProcessInfo processInfo] arguments];
@@ -2177,10 +2183,13 @@ NSDictionary *GSPrivateDefaultLocale()
 	    {
 	      NSString	*old = nil;
 
-	      /* anything beginning with a '-' is a defaults key and we must strip
-	          the '-' from it.  As a special case, we leave the '- in place
-	          for '-GS...' and '--GS...' for backward compatibility. */
-	      if ([key hasPrefix: @"-GS"] == YES || [key hasPrefix: @"--GS"] == YES)
+	      /* anything beginning with a '-' is a defaults key and we
+               * must strip the '-' from it.
+               * As a special case, we leave the '- in place for '-GS...'
+               * and '--GS...' for backward compatibility.
+               */
+	      if ([key hasPrefix: @"-GS"] == YES
+                || [key hasPrefix: @"--GS"] == YES)
 	        {
 	          old = key;
 	        }
@@ -2196,7 +2205,8 @@ NSDictionary *GSPrivateDefaultLocale()
 	          done = YES;
 	          continue;
 	        }
-	      else if ([val hasPrefix: @"-"] == YES && [val isEqual: @"-"] == NO)
+	      else if ([val hasPrefix: @"-"] == YES
+                && [val isEqual: @"-"] == NO)
 	        {  // Yet another argument
 	          [argDict setObject: @"" forKey: key];		// arg is empty.
 	          if (old != nil)
@@ -2241,10 +2251,12 @@ NSDictionary *GSPrivateDefaultLocale()
 	    }
           done = ((key = [enumerator nextObject]) == nil);
         }
+      parsingArguments = NO;
       [_lock unlock];
     }
   NS_HANDLER
     {
+      parsingArguments = NO;
       [_lock unlock];
       [localException raise];
     }
@@ -2342,7 +2354,7 @@ static BOOL isLocked = NO;
   NSEnumerator		*enumerator;
   NSString		*domainName;
   NSFileManager		*mgr;
-  BOOL			haveNewDomain = NO;
+  BOOL			haveChange = NO;
 
   mgr = [NSFileManager defaultManager];
 
@@ -2379,7 +2391,7 @@ static BOOL isLocked = NO;
 			      owner: self];
 	      [_persDomains setObject: pd forKey: domainName];
 	      [pd release];
-	      haveNewDomain = YES;
+	      haveChange = YES;
 	    }
 	  if (YES == [_searchList containsObject: domainName])
 	    {
@@ -2387,11 +2399,14 @@ static BOOL isLocked = NO;
 	       * synchronize to load the domain contents into memory
 	       * so a lookup will work.
 	       */
-	      [pd synchronize];
+              if (YES == [pd synchronize])
+                {
+                  haveChange = YES;
+                }
 	    }
 	}
     }
-  return haveNewDomain;
+  return haveChange;
 }
 
 - (BOOL) _readOnly
@@ -2472,12 +2487,12 @@ static BOOL isLocked = NO;
 
 - (BOOL) synchronize
 {
-  BOOL		wasLocked;
-  BOOL		result;
+  BOOL  wasLocked;
+  BOOL	hadChange = NO; // Have we read a change from disk?
 
   if (NO == [owner _lockDefaultsFile: &wasLocked])
     {
-      result = NO;
+      hadChange = NO;
       wasLocked = NO;
     }
   else
@@ -2490,6 +2505,7 @@ static BOOL isLocked = NO;
       if (YES == modified && NO == [owner _readOnly])
 	{
 	  NSDate	*mod;
+          BOOL          result;
 
           mod = [NSDate date];
 	  if (0 == [contents count])
@@ -2545,15 +2561,15 @@ static BOOL isLocked = NO;
 			}
 		    }
 		}
+              hadChange = YES;
 	    }
-	  result = YES;
 	}
       if (NO == wasLocked)
 	{
 	  [owner _unlockDefaultsFile];
 	}
     }
-  return result;
+  return hadChange;
 }
 
 - (NSDate*) updated
