@@ -6,6 +6,20 @@
 
 #import	"config.h"
 
+#if	defined(HAVE_STRING_H)
+/* For POSIX strerror_r() and others
+ */
+#include <string.h>
+#endif
+
+#if	defined(HAVE_STRINGS_H)
+/* For strcasecmp() and others
+ */
+#include <strings.h>
+#endif
+
+#include <errno.h>
+
 /* If this is included in a file in the Additions subdirectory, and we are
  * building for use with the NeXT/Apple Foundation, then we need to import
  * the native headers in preference to any of our own.
@@ -15,6 +29,7 @@
 #endif
 
 #import	"GNUstepBase/GSConfig.h"
+#import	"GNUstepBase/GSVersionMacros.h"
 
 /* Set localisation macro for use within the base library itsself.
  */
@@ -34,6 +49,7 @@
  * "Foundation/NSObject.h" first, to ensure we have a local copy.
  */
 #import	"Foundation/NSObject.h"
+#import	"GNUstepBase/NSObject+GNUstepBase.h"
 
 /* These headers are used in almost every file.
  */
